# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import eye, lil_matrix

from ._helpers import Assemble, Results


class MultiPointConstraint:
    """A Multi-point-constraint which connects a center-point to a list of points.

    Parameters
    ----------
    field : FieldContainer
        A field container with the displacement field as first field.
    points : (n,) ndarray
        An array with indices of points to be connected to the center-point.
    centerpoint : int
        The index of the centerpoint.
    skip : 3-tuple of bool, optional
        A tuple with boolean values for each axis to skip. If True, the respective axis
        is not connected. Default is (False, False, False).
    multiplier : float, optional
        A multiplier to penalize the relative displacements between the center-point and
        the points. Default is 1e-3.

    Notes
    -----
    A :class:`~felupe.MultiPointConstraint` is supported as an item in a
    :class:`~felupe.Step`. It provides the assemble-methods
    :meth:`MultiPointConstraint.assemble.vector() <felupe.MultiPointConstraint.assemble.vector>`
    and :meth:`MultiPointConstraint.assemble.matrix() <felupe.MultiPointConstraint.assemble.matrix>`.

    ..  note::

        Rotational degrees-of-freedom of the center-point are not connected to the
        points.

    Examples
    --------
    This example shows how to use a :class:`~felupe.MultiPointConstraint`.

    An additional center-point is added to a mesh. By default, all *hanging* points are
    collected in the mesh-attribute
    :attr:`Mesh.points_without_cells <felupe.Mesh.points_without_cells>`. The degrees of
    freedom of these points are considered as fixed, i.e. they are ignored. The center-
    point is not connected to any cell and is added to the points-without-cells array
    on :meth:`Mesh.update <felupe.Mesh.update>`. Hence, center-point has to be removed
    manually.

    ..  pyvista-plot::
        :context:

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=3)
        >>> mesh.update(points=np.vstack([mesh.points, [2.0, 0.5, 0.5]]))
        >>>
        >>> # prevent the field-values at the center-point to be treated as dof0
        >>> mesh.points_without_cells = mesh.points_without_cells[:-1]
        >>>
        >>> region = fem.RegionHexahedron(mesh)
        >>> displacement = fem.Field(region, dim=3)
        >>> field = fem.FieldContainer([displacement])
        >>>
        >>> umat = fem.NeoHooke(mu=1.0, bulk=2.0)
        >>> solid = fem.SolidBody(umat=umat, field=field)

    A :class:`~felupe.MultiPointConstraint` defines the multi-point constraint which
    connects the displacement degrees of freedom of the center-point with the dofs of
    points located at :math:`x=1`.

    ..  pyvista-plot::
        :context:
        :force_static:

        >>> import pyvista as pv
        >>>
        >>> mpc = fem.MultiPointConstraint(
        ...     field=field,
        ...     points=np.arange(mesh.npoints)[mesh.x == 1],
        ...     centerpoint=-1,
        ... )
        >>>
        >>> plotter = pv.Plotter()
        >>> actor_1 = plotter.add_points(
        ...     mesh.points[mpc.points],
        ...     point_size=16,
        ...     color="red",
        ... )
        >>> actor_2 = plotter.add_points(
        ...     mesh.points[[mpc.centerpoint]],
        ...     point_size=16,
        ...     color="green",
        ... )
        >>> mesh.plot(plotter=mpc.plot(plotter=plotter)).show()

    The mesh is fixed on the left end face and a ramped :class:`~felupe.PointLoad` is
    applied on the center-point of the :class:`~felupe.MultiPointConstraint`. All items
    are added to a :class:`~felupe.Step` and a :class:`~felupe.Job` is evaluated.

    ..  pyvista-plot::
        :context:

        >>> boundaries = {"fixed": fem.Boundary(displacement, fx=0)}
        >>> load = fem.PointLoad(field, points=[-1])
        >>> table = fem.math.linsteps([0, 1], num=5, axis=0, axes=3)
        >>>
        >>> step = fem.Step(
        ...     [solid, mpc, load], boundaries=boundaries, ramp={load: table}
        ... )
        >>> job = fem.Job([step]).evaluate()

    A view on the deformed mesh including the :class:`~felupe.MultiPointConstraint` is
    plotted.

    ..  pyvista-plot::
        :context:
        :force_static:

        >>> plotter = pv.Plotter()
        >>>
        >>> actor_1 = plotter.add_points(
        ...     mesh.points[mpc.points] + displacement.values[mpc.points],
        ...     point_size=16,
        ...     color="red",
        ... )
        >>> actor_2 = plotter.add_points(
        ...     mesh.points[[mpc.centerpoint]] + displacement.values[[mpc.centerpoint]],
        ...     point_size=16,
        ...     color="green",
        ... )
        >>> field.plot("Displacement", component=None, plotter=mpc.plot(plotter=plotter)).show()

    See Also
    --------
    felupe.MultiPointContact : A frictionless point-to-rigid (wall) contact.

    """

    def __init__(
        self, field, points, centerpoint, skip=(False, False, False), multiplier=1e3
    ):
        self.field = field
        self.mesh = field.region.mesh
        self.points = np.asarray(points)
        self.centerpoint = centerpoint
        self.mask = ~np.array(skip, dtype=bool)[: self.mesh.dim]
        self.axes = np.arange(self.mesh.dim)[self.mask]
        self.multiplier = multiplier

        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

    def plot(self, plotter=None, color="black", deformed=True, **kwargs):
        import pyvista as pv

        if plotter is None:
            plotter = pv.Plotter()

        # get undeformed points
        x = self.mesh.points

        if deformed:  # get deformed points
            x = x + self.field[0].values

        x = np.pad(x, ((0, 0), (0, 3 - x.shape[1])))
        pointa = x[self.centerpoint]

        for pointb in x[self.points]:
            plotter.add_mesh(pv.Line(pointa, pointb), color=color, **kwargs)

        return plotter

    def _vector(self, field=None, parallel=False):
        "Calculate vector of residuals with RBE2 contributions."

        if field is not None:
            self.field = field

        u = self.field.fields[0].values
        N = self.multiplier * (-u[self.points] + u[self.centerpoint])
        N[:, ~self.mask] = 0

        r = lil_matrix(u.shape)
        r[self.points] = -N
        r[self.centerpoint] = N.sum(axis=0)

        self.results.force = r.reshape(-1, 1).tocsr()
        return self.results.force

    def _matrix(self, field=None, parallel=False):
        "Calculate stiffness with RBE2 contributions."

        if field is not None:
            self.field = field

        indices = np.arange(self.mesh.ndof).reshape(self.mesh.points.shape)
        td = [indices[self.points.reshape(-1, 1), ax].ravel() for ax in self.axes]
        cd = [indices[self.centerpoint, ax].ravel() for ax in self.axes]

        L = lil_matrix((self.mesh.ndof, self.mesh.ndof))

        for t, c in zip(td, cd):
            L[t.reshape(-1, 1), t] = eye(len(t)) * self.multiplier
            L[t.reshape(-1, 1), c] = -self.multiplier
            L[c.reshape(-1, 1), t] = -self.multiplier
            L[c.reshape(-1, 1), c] = eye(len(c)) * self.multiplier * len(self.points)

        self.results.stiffness = L.tocsr()
        return self.results.stiffness


class MultiPointContact:
    """A frictionless point-to-rigid (wall) contact which connects a center-point to a
    list of points.

    Parameters
    ----------
    field : FieldContainer
        A field container with the displacement field as first field.
    points : (n,) ndarray
        An array with indices of points to be connected to the center-point.
    centerpoint : int
        The index of the centerpoint.
    skip : 3-tuple of bool, optional
        A tuple with boolean values for each axis to skip. If True, the respective axis
        is not connected. Default is (False, False, False).
    multiplier : float, optional
        A multiplier to penalize the relative displacements between the center-point and
        the points in contact. Default is 1e-6.

    Notes
    -----
    A :class:`~felupe.MultiPointContact` is supported as an item in a
    :class:`~felupe.Step`. It provides the assemble-methods
    :meth:`MultiPointContact.assemble.vector() <felupe.MultiPointContact.assemble.vector>`
    and :meth:`MultiPointContact.assemble.matrix() <felupe.MultiPointContact.assemble.matrix>`.

    Examples
    --------
    This example shows how to use a :class:`~felupe.MultiPointContact`.

    An additional center-point is added to a mesh. By default, all *hanging* points are
    collected in the mesh-attribute
    :attr:`Mesh.points_without_cells <felupe.Mesh.points_without_cells>`. The degrees of
    freedom of these points are considered as fixed, i.e. they are ignored. The center-
    point is not connected to any cell and is added to the points-without-cells array
    on :meth:`Mesh.update <felupe.Mesh.update>`. Hence, center-point has to be removed
    manually.

    ..  pyvista-plot::
        :context:

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=3)
        >>> mesh.update(points=np.vstack([mesh.points, [2.0, 0.5, 0.5]]))
        >>>
        >>> # prevent the field-values at the center-point to be treated as dof0
        >>> mesh.points_without_cells = mesh.points_without_cells[:-1]
        >>>
        >>> region = fem.RegionHexahedron(mesh)
        >>> displacement = fem.Field(region, dim=3)
        >>> field = fem.FieldContainer([displacement])
        >>>
        >>> umat = fem.NeoHooke(mu=1.0, bulk=2.0)
        >>> solid = fem.SolidBody(umat=umat, field=field)

    A :class:`~felupe.MultiPointContact` defines the multi-point contact which
    connects the displacement degrees of freedom of the center-point with the dofs of
    points located at :math:`x=1` if they are in contact. Only the :math:`x`-component
    is considered in this example.

    ..  pyvista-plot::
        :context:
        :force_static:

        >>> import pyvista as pv
        >>>
        >>> contact = fem.MultiPointContact(
        ...     field=field,
        ...     points=np.arange(mesh.npoints)[mesh.x == 1],
        ...     centerpoint=-1,
        ...     skip=(False, True, True)
        ... )
        >>>
        >>> plotter = pv.Plotter()
        >>> actor_1 = plotter.add_points(
        ...     mesh.points[mpc.points],
        ...     point_size=16,
        ...     color="red",
        ... )
        >>> actor_2 = plotter.add_points(
        ...     mesh.points[[mpc.centerpoint]],
        ...     point_size=16,
        ...     color="green",
        ... )
        >>> mesh.plot(plotter=contact.plot(plotter=plotter)).show()

    The mesh is fixed on the left end face and a ramped :class:`~felupe.Boundary` is
    applied on the center-point of the :class:`~felupe.MultiPointContact`. All items
    are added to a :class:`~felupe.Step` and a :class:`~felupe.Job` is evaluated.

    ..  pyvista-plot::
        :context:

        >>> boundaries = {
        ...     "fixed": fem.Boundary(displacement, fx=0),
        ...     "control": fem.Boundary(displacement, fx=2, skip=(1, 0, 0)),
        ...     "move": fem.Boundary(displacement, fx=2, skip=(0, 1, 1)),
        ... }
        >>> table = fem.math.linsteps([0, -1, -1.5], num=5)
        >>> step = fem.Step(
        ...     [solid, contact],
        ...     boundaries=boundaries,
        ...     ramp={boundaries["move"]: table},
        ... )
        >>> job = fem.Job([step]).evaluate()

    A view on the deformed mesh including the :class:`~felupe.MultiPointContact` is
    plotted.

    ..  pyvista-plot::
        :context:
        :force_static:

        >>> plotter = pv.Plotter()
        >>>
        >>> actor_1 = plotter.add_points(
        ...     mesh.points[contact.points] + displacement.values[contact.points],
        ...     point_size=16,
        ...     color="red",
        ... )
        >>> actor_2 = plotter.add_points(
        ...     mesh.points[[contact.centerpoint]] + displacement.values[[contact.centerpoint]],
        ...     point_size=16,
        ...     color="green",
        ... )
        >>> field.plot("Displacement", component=None, plotter=contact.plot(plotter=plotter)).show()

    See Also
    --------
    felupe.MultiPointConstraint : A Multi-point-constraint which connects a center-point
        to a list of points.

    """

    def __init__(
        self, field, points, centerpoint, skip=(False, False, False), multiplier=1e6
    ):
        self.field = field
        self.mesh = field.region.mesh
        self.points = np.asarray(points)
        self.centerpoint = centerpoint
        self.mask = ~np.array(skip, dtype=bool)[: self.mesh.dim]
        self.axes = np.arange(self.mesh.dim)[self.mask]
        self.multiplier = multiplier

        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

    def plot(
        self,
        plotter=None,
        offset=0,
        show_edges=True,
        color="black",
        opacity=0.5,
        deformed=True,
        **kwargs,
    ):
        import pyvista as pv

        if plotter is None:
            plotter = pv.Plotter()

        # get edge lengths of undeformed enclosing box
        x = self.mesh.points

        if deformed:  # get deformed points
            x = x + self.field[0].values

        edges = np.diag((x.max(axis=0) - x.min(axis=0))) + x.min(axis=0)

        # plot a line or a rectangle for each active contact plane
        for ax in self.axes:
            # fill the point values of the normal axis with the centerpoint values
            points = edges.copy()
            points[:, ax] = x[self.centerpoint, ax] + offset

            # scale the line or rectangle at the origin
            origin = points.mean(axis=0)
            points = (points - origin) * 1.05 + origin

            # plot a line or a rectangle
            if len(points) == 3:
                plotter.add_mesh(
                    pv.Rectangle(points), color=color, opacity=opacity, **kwargs
                )
            else:
                points = np.pad(points, ((0, 0), (0, 3 - points.shape[1])))
                plotter.add_mesh(
                    pv.Line(*points), color=color, opacity=opacity, **kwargs
                )

        return plotter

    def _vector(self, field=None, parallel=False):
        "Calculate vector of residuals with RBE2 contributions."

        if field is not None:
            self.field = field

        u = self.field.fields[0].values

        Xc = self.mesh.points[self.centerpoint]
        Xt = self.mesh.points[self.points]

        xc = u[self.centerpoint] + Xc
        xt = u[self.points] + Xt

        mask = np.sign(-Xt + Xc) == np.sign(-xt + xc)
        mask[:, ~self.mask] = True
        n = -xt + xc
        n[mask] = 0

        r = lil_matrix(u.shape)
        r[self.points] = -self.multiplier * n
        r[self.centerpoint] = self.multiplier * n.sum(axis=0)

        self.results.force = r.reshape(-1, 1).tocsr()
        return self.results.force

    def _matrix(self, field=None, parallel=False):
        "Calculate stiffness with RBE2 contributions."

        if field is not None:
            self.field = field

        u = self.field.fields[0].values

        Xc = self.mesh.points[self.centerpoint]
        Xt = self.mesh.points[self.points]

        xc = u[self.centerpoint] + Xc
        xt = u[self.points] + Xt

        mask = np.sign(-Xt + Xc) != np.sign(-xt + xc)
        masks = [mask[:, ax] for ax in self.axes]

        indices = np.arange(self.mesh.ndof).reshape(self.mesh.points.shape)
        td = [indices[self.points.reshape(-1, 1), ax].ravel() for ax in self.axes]
        cd = [indices[self.centerpoint, ax].ravel() for ax in self.axes]

        L = lil_matrix((self.mesh.ndof, self.mesh.ndof))

        for t, c, m in zip(td, cd, masks):
            L[t[m].reshape(-1, 1), t[m]] = eye(len(t[m])) * self.multiplier
            L[t[m].reshape(-1, 1), c] = -self.multiplier
            L[c.reshape(-1, 1), t[m]] = -self.multiplier
            L[c.reshape(-1, 1), c] = eye(len(c)) * self.multiplier * len(self.points[m])

        self.results.stiffness = L.tocsr()
        return self.results.stiffness



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from jax.numpy import trace
from jax.numpy.linalg import det

from ....tensortrax.models.hyperelastic import neo_hooke as neo_hooke_docstring


@wraps(neo_hooke_docstring)
def neo_hooke(C, mu):
    return mu / 2 * (det(C) ** (-1 / 3) * trace(C) - 3)



from numpy import sqrt

from ._field import (
    deformation_gradient,
    displacement,
    extract,
    grad,
    hess,
    interpolate,
    norm,
    right_cauchy_green_deformation,
    strain,
    strain_stretch_1d,
    values,
)
from ._math import linsteps
from ._solve import solve_2d, solve_nd
from ._spatial import revolve_points, rotate_points, rotation_matrix
from ._tensor import (
    cdya,
    cdya_ik,
    cdya_il,
    cof,
    cross,
    dddot,
    ddot,
    det,
    dev,
    dot,
    dya,
    eig,
    eigh,
    eigvals,
    eigvalsh,
    equivalent_von_mises,
    identity,
    inplane,
    inv,
    majortranspose,
    ravel,
    reshape,
    svd,
    sym,
    tovoigt,
    trace,
    transpose,
)

__all__ = [
    "sqrt",
    "extract",
    "grad",
    "hess",
    "interpolate",
    "norm",
    "strain",
    "strain_stretch_1d",
    "values",
    "linsteps",
    "rotation_matrix",
    "cdya",
    "cdya_ik",
    "cdya_il",
    "cof",
    "cross",
    "ddot",
    "dddot",
    "deformation_gradient",
    "det",
    "dev",
    "displacement",
    "dot",
    "dya",
    "eig",
    "eigh",
    "eigvals",
    "eigvalsh",
    "equivalent_von_mises",
    "identity",
    "inplane",
    "inv",
    "majortranspose",
    "ravel",
    "reshape",
    "right_cauchy_green_deformation",
    "svd",
    "sym",
    "tovoigt",
    "trace",
    "transpose",
    "solve_2d",
    "solve_nd",
    "revolve_points",
    "rotate_points",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import sum as tsum
from tensortrax.math.linalg import det, eigvalsh


def storakers(C, mu, alpha, beta):
    r"""Strain energy function of the Storåkers isotropic hyperelastic
    `foam <https://doi.org/10.1016/0022-5096(86)90033-5>`_ material formulation [1]_.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    mu : list of float
        List of moduli.
    alpha : list of float
        List of stretch exponents.
    beta : list of float
        List of coefficients for the degree of compressibility.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-foam`

    ..  math::
        :label: psi-foam

        \psi = \sum_i \frac{2 \mu_i}{\alpha^2_i} \left[
            \lambda_1^{\alpha_i} +
            \lambda_2^{\alpha_i} +
            \lambda_3^{\alpha_i} - 3
            + \frac{1}{\beta_i} \left( J^{-\alpha_i \beta_i} - 1 \right)
        \right]

    The sum of the moduli :math:`\mu_i` is equal to the initial shear modulus
    :math:`\mu`, see Eq. :eq:`shear-modulus-foam`,

    ..  math::
        :label: shear-modulus-foam

        \mu = \sum_i \mu_i

    and the initial bulk modulus is given in Eq. :eq:`bulk-modulus-foam`.

    ..  math::
        :label: bulk-modulus-foam

        K = \sum_i 2 \mu_i \left( \frac{1}{3} + \beta_i \right)

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> import felupe as fem
        >>>
        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.storakers,
        ...     mu=[4.5 * (1.85 / 2), -4.5 * (-9.2 / 2)],
        ...     alpha=[1.85, -9.2],
        ...     beta=[0.92, 0.92],
        ... )
        >>> ax = umat.plot(
        ...     ux=fem.math.linsteps([1, 2], 15),
        ...     ps=fem.math.linsteps([1, 1], 15),
        ...     bx=fem.math.linsteps([1, 1], 9),
        ... )

    References
    ----------
    .. [1] B. Storåkers, "On material representation and constitutive branching in
       finite compressible elasticity", Journal of the Mechanics and Physics of Solids,
       vol. 34, no. 2. Elsevier BV, pp. 125–145, Jan. 1986. doi:
       10.1016/0022-5096(86)90033-5.
    """

    λ2 = eigvalsh(C)

    return tsum(
        [
            2 * μ / α**2 * (tsum(λ2 ** (α / 2)) - 3 + (det(C) ** (-α * β / 2) - 1) / β)
            for μ, α, β in zip(mu, alpha, beta)
        ]
    )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import trace
from tensortrax.math.linalg import det


def mooney_rivlin(C, C10, C01):
    r"""Strain energy function of the isotropic hyperelastic
    `Mooney-Rivlin <https://en.wikipedia.org/wiki/Mooney-Rivlin_solid>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    C10 : float
        First material parameter associated to the first invariant.
    C01 : float
        Second material parameter associated to the second invariant.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-mr`

    ..  math::
        :label: psi-mr

        \psi = C_{10} \left(\hat{I}_1 - 3 \right) + C_{01} \left(\hat{I}_2 - 3 \right)

    with the first and second main invariant of the distortional part of the right
    Cauchy-Green deformation tensor, see Eq. :eq:`invariants-mr`.

    ..  math::
        :label: invariants-mr

        \hat{I}_1 &= J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

        \hat{I}_2 &= J^{-4/3} \frac{1}{2} \left(
            \text{tr}\left(\boldsymbol{C}\right)^2 -
            \text{tr}\left(\boldsymbol{C}^2\right)
        \right)

    The doubled sum of both material parameters is equal to the shear modulus
    :math:`\mu` as denoted in Eq. :eq:`shear-modulus-mr`.

    ..  math::
        :label: shear-modulus-mr

        \mu = 2 \left( C_{10} + C_{01} \right)

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.mooney_rivlin, C10=0.3, C01=0.8
        ... )
        >>> ax = umat.plot(incompressible=True)

    """
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = (I1**2 - J3**2 * trace(C @ C)) / 2
    return C10 * (I1 - 3) + C01 * (I2 - 3)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


def interpolate_line(mesh, xi, axis=None, Interpolator=None, **kwargs):
    r"""Return an interpolated line mesh from an existing line mesh with provided
    interpolation points on a given axis.

    Parameters
    ----------
    mesh : Mesh
        A :class:`~felupe.Mesh` with cell-type ``line``.
    xi : ndarray
        Points at which to interpolate data. If axis is None, a normalized curve-
        progress must be provided (:math:`0 \le \xi \le 1`).
    axis : int or None, optional
        The axis of the points at which the data will be interpolated. If None, a
        normalized curve-progress is used. Default is None.
    Interpolator : callable or None, optional
        An interpolator class (default is :class:`scipy.interpolate.PchipInterpolator`).
    **kwargs : dict, optional
        Optional keyword arguments are passed to the Interpolator.

    Returns
    -------
    Mesh
        A new line mesh with interpolated points. The attribute ``points_derivative``
        holds the derivatives of the independent variable w.r.t. the dependent
        variable(s).

    Examples
    --------
    ..  pyvista-plot::
        :context:
        :force_static:

        >>> import felupe as fem
        >>> import numpy as np
        >>> from scipy.interpolate import CubicSpline
        >>>
        >>> mesh = fem.mesh.Line(b=1.0, n=5).expand(n=1)
        >>> t = mesh.x.copy()
        >>> mesh.points[:, 0] = np.sin(2 * np.pi * t)
        >>> mesh.points[:, 1] = np.cos(2 * np.pi * t)
        >>>
        >>> mesh_new = fem.mesh.interpolate_line(
        ...     mesh, xi=np.linspace(0, 1), Interpolator=CubicSpline, bc_type="periodic"
        ... )
        >>>
        >>> mesh_new.plot(
        ...     plotter=mesh.plot(style="points", color="red", point_size=15),
        ...     color="black",
        ... ).show()

    """

    if Interpolator is None:
        from scipy.interpolate import PchipInterpolator as Interpolator

    if axis is None:  # progress-based interpolation

        distances = np.linalg.norm(np.diff(mesh.points, axis=0), axis=1)
        progress = np.insert(np.cumsum(distances), 0, 0)
        progress /= progress.max()

        spline = Interpolator(progress, mesh.points, **kwargs)

        points_new = spline(xi)
        cells_new = np.repeat(np.arange(len(xi)), 2)[1:-1].reshape(-1, 2)

        mesh_new = type(mesh)(points_new, cells_new, cell_type="line")
        mesh_new.points_derivative = spline.derivative()(xi)

    else:  # independent- and dependent-variables

        # line connectivity
        # concatenate the first point of each cell and the last point of the last cell
        line = np.concatenate([mesh.cells[:, :-1].ravel(), mesh.cells[-1, -1:]])

        # independent spline variable
        points = mesh.points[line, axis]
        ascending = np.argsort(points)

        # dependent spline variable(s)
        mask = np.ones(mesh.dim, dtype=bool)
        mask[axis] = False

        axes = np.arange(mesh.dim)
        values = mesh.points[line.reshape(-1, 1), axes[mask]]

        # create a spline
        spline = Interpolator(points[ascending], values[ascending], **kwargs)

        # evaluation points for the independent spline variable
        points_new = np.zeros((len(xi), mesh.dim))
        points_new[:, axis] = xi
        points_new[:, axes[mask]] = spline(xi)

        cells_new = np.repeat(np.arange(len(xi)), 2)[1:-1].reshape(-1, 2)

        mesh_new = type(mesh)(points_new, cells_new, cell_type="line")
        mesh_new.points_derivative = np.zeros_like(mesh_new.points)
        mesh_new.points_derivative[:, axes[mask]] = spline.derivative()(xi)

    return mesh_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from .._integral import IntegralForm
from ._bilinear import BilinearForm
from ._linear import LinearForm


class LinearFormExpression:
    r"""A linear form object with methods for integration and assembly of
    vectors where ``v`` is a basis object on a field container.

    ..  math::

        L(v) = \int_\Omega f \cdot v \ dx

    Parameters
    ----------
    v : Basis
        Basis (shape) functions and gradients of a field.


    """

    def __init__(self, v):
        self.v = v
        self.dx = self.v.field[0].region.dV
        self._form = IntegralForm(np.zeros(len(v.field.fields)), self.v.field, self.dx)

        self._linearform = [LinearForm(v=vi, dx=self.dx) for vi in self.v]

    def integrate(self, weakform, kwargs=None, parallel=False):
        r"""Return evaluated (but not assembled) integrals.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, **kwargs)``.
        kwargs : dict or None, optional
            Optional named arguments for callable weakform (default is None).
        parallel : bool, optional
            Flag to activate parallel threading (default is False).

        Returns
        -------
        values : ndarray
            Integrated (but not assembled) vector values.
        """

        if kwargs is None:
            kwargs = {}

        return [
            form.integrate(fun, kwargs, parallel=parallel)
            for form, fun in zip(self._linearform, weakform)
        ]

    def assemble(self, weakform, kwargs=None, parallel=False):
        r"""Return the assembled integral as vector.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, **kwargs)``.
        kwargs : dict or None, optional
            Optional named arguments for callable weakform (default is None).
        parallel : bool, optional
            Flag to activate parallel threading (default is False).

        Returns
        -------
        values : csr_matrix
            The assembled vector.
        """

        values = self.integrate(weakform, kwargs, parallel=parallel)

        return self._form.assemble(values)


class BilinearFormExpression:
    r"""A bilinear form object with methods for integration and assembly of
    matrices where ``v`` is a mixed-basis object on mixed-fields.

    ..  math::

        a(v, u) = \int_\Omega v \cdot f \cdot u \ dx

    Parameters
    ----------
    v : Basis
        Basis (shape) functions and gradients of a field.
    u : Basis
        Basis (shape) functions and gradients of a field.

    """

    def __init__(self, v, u):
        self.v = v
        self.u = u
        self.dx = self.v.field[0].region.dV

        self.nv = len(v.field.fields)
        self.i, self.j = np.triu_indices(self.nv)

        self._form = IntegralForm(
            fun=np.zeros(len(self.i)),
            v=self.v.field,
            dV=self.dx,
            u=self.u.field,
        )

        self._bilinearform = []

        for a, (i, j) in enumerate(zip(self.i, self.j)):
            self._bilinearform.append(
                BilinearForm(
                    v=self.v[i],
                    u=self.u[j],
                    dx=self.dx,
                )
            )

    def integrate(self, weakform, kwargs=None, parallel=False, sym=False):
        r"""Return evaluated (but not assembled) integrals.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, u, **kwargs)``.
        kwargs : dict or None, optional
            Optional named arguments for callable weakform (default is None).
        parallel : bool, optional (default is False)
            Flag to activate parallel threading.

        Returns
        -------
        values : ndarray
            Integrated (but not assembled) matrix values.
        """

        if kwargs is None:
            kwargs = {}

        return [
            form.integrate(fun, kwargs, parallel=parallel, sym=sym)
            for form, fun in zip(self._bilinearform, weakform)
        ]

    def assemble(self, weakform, kwargs=None, parallel=False, sym=False):
        r"""Return the assembled integral as matrix.

        Parameters
        ----------
        weakform : callable
            A callable function ``weakform(v, u, **kwargs)``.
        kawargs : dict or None, optional
            Optional named arguments for callable weakform (default is None).
        parallel : bool, optional
            Flag to activate parallel threading (default is False).

        Returns
        -------
        values : csr_matrix
            The assembled sparse matrix.
        """

        values = self.integrate(weakform, kwargs, parallel=parallel, sym=sym)

        return self._form.assemble(values)



from ._base import CompositeMaterial, ConstitutiveMaterial, constitutive_material
from ._kinematics import AreaChange, LineChange, VolumeChange
from ._material import Material
from ._mixed import NearlyIncompressible, ThreeFieldVariation
from ._view import ViewMaterial, ViewMaterialIncompressible
from .hyperelasticity import NeoHooke, NeoHookeCompressible, OgdenRoxburgh, Volumetric
from .linear_elasticity import (
    LinearElastic,
    LinearElastic1D,
    LinearElasticLargeStrain,
    LinearElasticOrthotropic,
    LinearElasticPlaneStrain,
    LinearElasticPlaneStress,
    LinearElasticTensorNotation,
    lame_converter,
    lame_converter_orthotropic,
)
from .poisson import Laplace
from .strain import MaterialStrain
from .strain.models import (
    LinearElasticPlasticIsotropicHardening,
    linear_elastic,
    linear_elastic_plastic_isotropic_hardening,
    linear_elastic_viscoelastic,
)

__all__ = [
    "NeoHooke",
    "NeoHookeCompressible",
    "Laplace",
    "LinearElastic",
    "LinearElastic1D",
    "LinearElasticLargeStrain",
    "LinearElasticOrthotropic",
    "LinearElasticPlaneStrain",
    "LinearElasticPlaneStress",
    "LinearElasticTensorNotation",
    "lame_converter",
    "lame_converter_orthotropic",
    "OgdenRoxburgh",
    "LinearElasticPlasticIsotropicHardening",
    "Material",
    "MaterialStrain",
    "AreaChange",
    "LineChange",
    "VolumeChange",
    "ThreeFieldVariation",
    "NearlyIncompressible",
    "linear_elastic",
    "linear_elastic_viscoelastic",
    "linear_elastic_plastic_isotropic_hardening",
    "ViewMaterial",
    "ViewMaterialIncompressible",
    "ConstitutiveMaterial",
    "constitutive_material",
    "CompositeMaterial",
    "Volumetric",
]
try:
    from .tensortrax import Hyperelastic
    from .tensortrax import Material as MaterialAD
    from .tensortrax import isochoric_volumetric_split, total_lagrange, updated_lagrange
    from .tensortrax.models.hyperelastic import (
        alexander,
        anssari_benam_bucchi,
        arruda_boyce,
        blatz_ko,
        extended_tube,
        finite_strain_viscoelastic,
        lopez_pamies,
        miehe_goektepe_lulei,
        mooney_rivlin,
        neo_hooke,
        ogden,
        ogden_roxburgh,
        saint_venant_kirchhoff,
        saint_venant_kirchhoff_orthotropic,
        storakers,
        third_order_deformation,
        van_der_waals,
        yeoh,
    )
    from .tensortrax.models.lagrange import morph, morph_representative_directions

    __all_tensortrax = [
        "alexander",
        "anssari_benam_bucchi",
        "arruda_boyce",
        "blatz_ko",
        "extended_tube",
        "finite_strain_viscoelastic",
        "isochoric_volumetric_split",
        "lopez_pamies",
        "miehe_goektepe_lulei",
        "mooney_rivlin",
        "morph",
        "morph_representative_directions",
        "neo_hooke",
        "ogden",
        "ogden_roxburgh",
        "saint_venant_kirchhoff",
        "saint_venant_kirchhoff_orthotropic",
        "storakers",
        "third_order_deformation",
        "van_der_waals",
        "yeoh",
        "MaterialAD",
        "Hyperelastic",
        "total_lagrange",
        "updated_lagrange",
    ]
    __all__.extend(__all_tensortrax)

except ImportError:
    pass



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import eigsh

from ..dof import partition


class FreeVibration:
    """A Free-Vibration Step/Job.

    Parameters
    ----------
    items : list of SolidBody or SolidBodyNearlyIncompressible
        A list of items with methods for the assembly of sparse stiffness and mass
        matrices.
    boundaries : dict of Boundary, optional
        A dict with :class:`~felupe.Boundary` conditions (default is None).

    Notes
    -----
    ..  note::

        Boundary conditions with non-zero values are not supported.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>> import numpy as np
        >>>
        >>> mesh = fem.Rectangle(b=(5, 1), n=(50, 10))
        >>> region = fem.RegionQuad(mesh)
        >>> field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])
        >>>
        >>> boundaries = dict(left=fem.Boundary(field[0], fx=0))
        >>> solid = fem.SolidBody(fem.LinearElastic(E=2.5, nu=0.25), field, density=1.0)
        >>> modal = fem.FreeVibration(items=[solid], boundaries=boundaries).evaluate()
        >>>
        >>> eigenvector, frequency = modal.extract(n=4, inplace=True)
        >>> solid.plot("Stress", component=0).show()
    """

    def __init__(self, items, boundaries=None):
        self.items = items

        if boundaries is None:
            boundaries = {}

        self.boundaries = boundaries
        self.eigenvalues = None
        self.eigenvectors = None

    def evaluate(self, x0=None, solver=eigsh, parallel=False, **kwargs):
        if x0 is not None:
            x = x0
        else:
            x = self.items[0].field

        # link field of items with global field
        [item.field.link(x) for item in self.items]

        # init matrix with shape from global field
        shape = (np.sum(x.fieldsizes), np.sum(x.fieldsizes))
        stiffness = csr_matrix(shape)
        mass = csr_matrix(shape)

        # assemble matrices
        for item in self.items:
            K = item.assemble.matrix(parallel=parallel)
            M = item.assemble.mass()

            if item.assemble.multiplier is not None:
                K *= item.assemble.multiplier

            # check and reshape matrices
            if K.shape != stiffness.shape:
                K.resize(*shape)

            if M.shape != mass.shape:
                M.resize(*shape)

            # add matrices
            stiffness += K
            mass += M

        dof0, self.dof1 = partition(x, self.boundaries)

        K = stiffness[self.dof1][:, self.dof1]
        M = mass[self.dof1][:, self.dof1]

        sigma = kwargs.get("sigma", 0)
        self.eigenvalues, self.eigenvectors = solver(A=K, M=M, sigma=sigma, **kwargs)

        return self

    def extract(self, n=0, x0=None, inplace=True):
        if x0 is not None:
            field = x0
        else:
            field = self.items[0].field

        if not inplace:
            field = field.copy()

        values = np.zeros(sum(field.fieldsizes))
        values[self.dof1] = self.eigenvectors[:, n]

        frequency = np.sqrt(self.eigenvalues[n]) / (2 * np.pi)

        [f.fill(0) for f in field]
        field += values

        return field, frequency



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


class Scene:
    r"""Base class for plotting a static scene.

    Attributes
    ----------
    mesh : pyvista.UnstructuredGrid
        A generalized Dataset with the mesh as well as point- and cell-data. This is
        not an instance of :class:`felupe.Mesh`.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> scene = fem.view.Scene()
        >>> scene.mesh = fem.Cube(n=3).as_unstructured_grid()
        >>> scene.mesh.point_data["Displacement"] = np.arange(81).reshape(27, 3) / 300
        >>> scene.mesh.set_active_scalars(None)
        >>>
        >>> scene.plot("Displacement", component=None).show()

    See Also
    --------
    felupe.ViewMesh : Provide Visualization methods for a mesh with optional given
        dicts of point- and cell-data items.
    felupe.ViewField : Provide Visualization methods for a field container.
    felupe.ViewSolid : Provide Visualization methods for a field container or a
        solid body.
    """

    def plot(
        self,
        name=None,
        component=0,
        label=None,
        factor=1.0,
        show_edges=True,
        show_undeformed=True,
        cmap="turbo",
        view="default",
        theme="default",
        scalar_bar_args=None,
        scalar_bar_vertical=False,
        add_axes=True,
        off_screen=False,
        plotter=None,
        notebook=False,
        extract_surface=False,
        nonlinear_subdivision=1,
        smooth_shading=True,
        split_sharp_edges=True,
        edge_color="black",
        line_width=1.0,
        **kwargs,
    ):
        """Plot scalars, selected by name and component.

        Parameters
        ----------
        name: str or None, optional
            Name of array of scalars to plot (default is None).
        component : int, optional
            Component of vector-valued scalars to plot (default is 0).
        label : str or None, optional
            A custom label which is shown in the scalar bar. If no label is given, it is
            created by the name including the component. For vector-valued scalars, the
            component number is replaced by ``"X"``, etc. For 3d-tensors in
            full or reduced vector storage (Voigt-Notation) the component number is
            replaced by ``"XY"``, etc. If ``"Principal Values of"`` is in the name, the
            component number is replaced by ``"\n (Max. Principal)"``, assuming that the
            principal values are sorted in descending order.
        factor : float, optional
            Factor for the scaling of the warped (deformed) mesh (default is 1.0).
        show_edges : bool, optional
            Show the edges of the cells (default is True).
        show_undeformed : bool, optional
            Show the undeformed model (default is True).
        cmap : str, optional
            The color map (default is "turbo").
        view : str or None, optional
            The camera position, e.g. "xy" or "iso" (default is "default"). If not
            specified, this is None for 3d-meshes and "xy" for 2d-meshes.
        theme : str or None, optional
            The theme used for plotting, e.g. "default" or "document" (default is None).
        scalar_bar_vertical : bool, optional
            A flag to show the interactive scalar bar in vertical orientation on the
            right side (default is True).
        add_axes : bool, optional
            Add the axes, i.e. the coordinate system to the scene (default is True).
        off_screen : bool, optional
            Initialize the plotter off-screen and don't open a window on plotting. For
            screenshots, it is necessary to set ``off_screen=True`` (default is False).
        plotter : pyvista.Plotter or None, optional
            Use a given Plotter instead of creating a new instance (default is None).
        notebook : bool, optional
            When True, the resulting plot is placed inline a jupyter notebook. Assumes a
            jupyter console is active. Automatically enables off_screen (default is
            False).
        extract_surface : bool, optional
            Extract the surface mesh. Required to hide internal edges of quadratic
            cells (default is False). If True and ``show_edges=True``, the feature edges
            of a separated mesh are plotted.
        nonlinear_subdivision : int, optional
            Number of subdivisions to generate a smooth surface based on the mid-edge
            points (default is 1, no subdivision). If greater than 1, the surface of
            the mesh is extracted.
        smooth_shading : bool, optional
            A flag to enable smooth shading (default is True). Only considered if
            number of subdivisions is greater than 1.
        split_sharp_edges : bool, optional
            A flag to split sharp edges (default is True). Use this flag in combination
            with smooth shading. Only considered if number of subdivisions is greater
            than 1.
        edge_color : str, optional
            The color of the edge lines (default is "black").
        line_width : float, optional
            The line-width of the edge lines (default is 1.0).


        Returns
        -------
        plotter : pyvista.Plotter
            A Plotter object with methods ``plot()``, ``screenshot()``, etc.

        See Also
        --------
        pyvista.Plotter : Plotting object to display vtk meshes or numpy arrays.
        """

        import pyvista as pv

        if theme is not None:
            pv.set_plot_theme(theme)

        if plotter is None:
            plotter_kwargs = dict()
            if off_screen:
                plotter_kwargs["off_screen"] = off_screen
            if notebook:
                plotter_kwargs["notebook"] = notebook
            plotter = pv.Plotter(**plotter_kwargs)

        if scalar_bar_args is None:
            scalar_bar_args = {}

        if name is not None:
            if component is not None:
                if name in self.mesh.point_data.keys():
                    data = self.mesh.point_data[name]
                else:
                    data = self.mesh.cell_data[name]

        if name is not None and label is None:
            data_label = name

            if component is not None:
                if name in self.mesh.point_data.keys():
                    data = self.mesh.point_data[name]
                else:
                    data = self.mesh.cell_data[name]

                dim = 1

                if len(data.shape) == 2:
                    dim = data.shape[1]

                component_labels_dict = {
                    1: [""],
                    2: ["X", "Y"],
                    3: ["X", "Y", "Z"],
                    6: ["XX", "YY", "ZZ", "XY", "YZ", "XZ"],
                    9: [
                        "XX",
                        "XY",
                        "XZ",
                        "YX",
                        "YY",
                        "YZ",
                        "ZX",
                        "ZY",
                        "ZZ",
                    ],
                }

                if "Principal Values of " in data_label:
                    data[:] = np.flip(np.sort(data, axis=-1), axis=-1)
                    component_labels_dict[2] = [
                        "(Max. Principal)",
                        "(Min. Principal)",
                    ]
                    component_labels_dict[3] = [
                        "(Max. Principal)",
                        "(Int. Principal)",
                        "(Min. Principal)",
                    ]
                    data_label = data_label[20:]

                if "Equivalent of " in data_label:
                    component_labels_dict[1] = [""]

                component_labels = np.arange(dim)
                if dim in component_labels_dict.keys():
                    component_labels = component_labels_dict[dim]

                component_label = component_labels[component]

            else:
                component_label = "Magnitude"

            label = f"{data_label} {component_label}"

        if show_undeformed:
            plotter.add_mesh(
                self.mesh, show_edges=False, opacity=0.2, line_width=line_width
            )

        mesh = self.mesh
        if "Displacement" in self.mesh.point_data.keys():
            mesh = mesh.warp_by_vector("Displacement", factor=factor)

        surface = mesh
        show_edges_surface = show_edges
        kwargs_with_line_width = {**kwargs}

        if mesh.number_of_cells > 0:
            if extract_surface or nonlinear_subdivision > 1:
                surface = surface.extract_surface(
                    nonlinear_subdivision=nonlinear_subdivision
                )
                show_edges_surface = False
            else:
                kwargs_with_line_width["line_width"] = line_width

        # disable surface-related arguments if the mesh contains no cells
        if mesh.number_of_cells == 0 or nonlinear_subdivision == 1:
            smooth_shading = None
            split_sharp_edges = None

        # don't show edges for the base (surface) mesh to hide internal edges of
        # quadratic / Lagrange cell-types
        plotter.add_mesh(
            mesh=surface,
            scalars=name,
            component=component,
            show_edges=show_edges_surface,
            cmap=cmap,
            scalar_bar_args={
                "title": label,
                "interactive": True,
                "vertical": scalar_bar_vertical,
                **scalar_bar_args,
            },
            smooth_shading=smooth_shading,
            split_sharp_edges=split_sharp_edges,
            **kwargs_with_line_width,
        )

        # extract the feature edges (without cell-internal edges)
        if (
            mesh.number_of_cells > 0
            and show_edges
            and (extract_surface or nonlinear_subdivision > 1)
        ):
            edges = (
                mesh.separate_cells()
                .extract_surface(nonlinear_subdivision=nonlinear_subdivision)
                .extract_feature_edges()
            )
            actor = plotter.add_mesh(edges, color=edge_color, line_width=line_width)
            actor.mapper.SetResolveCoincidentTopologyToPolygonOffset()

        if view == "default":
            if np.allclose(self.mesh.points[:, 2], 0):
                view = "xy"
                plotter.enable_parallel_projection()

            else:
                view = None
                plotter.camera.elevation = -15
                plotter.camera.azimuth = -100

        plotter.camera_position = view
        # pv.set_plot_theme(theme)

        if add_axes:
            plotter.add_axes()

        return plotter



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from tensortrax.math import log, sqrt, trace
from tensortrax.math.linalg import det


def van_der_waals(C, mu, limit, a, beta):
    r"""Strain energy function of the
    `Van der Waals <https://doi.org/10.1016/0032-3861(81)90200-7>`_ [1]_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    mu : float
        Initial shear modulus.
    limit : float
        Limiting stretch :math:`\lambda_m` at which the polymer chain network becomes
        locked.
    a : float
        Attractive interactions between the quasi-particles.
    beta : float
        Mixed-Invariant factor: 0 for pure I1- and 1 for pure I2-contribution.

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.van_der_waals,
        ...     mu=1.0,
        ...     beta=0.1,
        ...     a=0.5,
        ...     limit=5.0
        ... )
        >>> ax = umat.plot(incompressible=True)

    References
    ----------
    ..  [1] H.-G. Kilian, "Equation of state of real networks", Polymer, vol. 22, no. 2.
        Elsevier BV, pp. 209–217, Feb. 1981. doi:
        `10.1016/0032-3861(81)90200-7 <https://www.doi.org/10.1016/0032-3861(81)90200-7>`_.

    """
    J3 = det(C) ** (-1 / 3)
    I1 = J3 * trace(C)
    I2 = (trace(C) ** 2 - J3**2 * trace(C @ C)) / 2
    Im = (1 - beta) * I1 + beta * I2
    Im += 1e-4
    eta = sqrt((Im - 3) / (limit**2 - 3))
    return mu * (
        -(limit**2 - 3) * (log(1 - eta) + eta) - 2 / 3 * a * ((Im - 3) / 2) ** (3 / 2)
    )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from jax.numpy import array, concatenate, diag, eye, maximum, sqrt, trace, triu_indices
from jax.numpy.linalg import det, eigvalsh, inv
from jax.scipy.linalg import expm

from ....tensortrax.models.lagrange import morph as morph_docstring
from ..._total_lagrange import total_lagrange


@wraps(morph_docstring)
@total_lagrange
def morph(F, statevars, p):
    # right Cauchy-Green deformation tensor
    C = F.T @ F

    # extract old state variables
    CTSn = statevars[0]
    from_triu = lambda C: C[array([[0, 1, 2], [1, 3, 4], [2, 4, 5]])]
    Cn = from_triu(statevars[1:7])
    SAn = from_triu(statevars[7:13])

    # distortional part of right Cauchy-Green deformation tensor
    I3 = det(C)
    CG = C * I3 ** (-1 / 3)

    # inverse of and incremental right Cauchy-Green deformation tensor
    invC = inv(C)
    dC = C - Cn

    # eigenvalues of right Cauchy-Green deformation tensor (sorted in ascending order)
    eigvalsh2 = lambda C: eigvalsh(C + diag(array([1e-4, -1e-4, 0])))
    λCG = eigvalsh2(CG)

    # Tresca invariant of distortional part of right Cauchy-Green deformation tensor
    CTG = λCG[-1] - λCG[0]

    # maximum Tresca invariant in load history
    CTS = maximum(CTG, CTSn)

    def sigmoid(x):
        "Algebraic sigmoid function."
        return 1 / sqrt(1 + x**2)

    # material parameters
    α = p[0] + p[1] * sigmoid(p[2] * CTS)
    β = p[3] * sigmoid(p[2] * CTS)
    γ = p[4] * CTS * (1 - sigmoid(CTS / p[5]))

    dev = lambda C: C - trace(C) / 3 * eye(3)
    sym = lambda C: (C + C.T) / 2

    LG = sym(dev(invC @ dC)) @ CG
    λLG = eigvalsh2(LG)
    LTG = λLG[-1] - λLG[0]

    # limiting stresses "L" and additional stresses "A"
    SL = (γ * expm(p[6] * LG / LTG * CTG / CTS) + p[7] * LG / LTG) @ invC
    SA = (SAn + β * LTG * SL) / (1 + β * LTG)

    # second Piola-Kirchhoff stress tensor
    S = 2 * α * dev(CG) @ invC + dev(SA @ C) @ invC

    i, j = triu_indices(3)
    to_triu = lambda C: C[i, j]
    statevars_new = concatenate([array([CTS]), to_triu(C), to_triu(SA)])

    return S, statevars_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ...math import cdya_ik, dot, identity, ravel, reshape, svd, sym, transpose
from .._base import ConstitutiveMaterial


class MaterialStrain(ConstitutiveMaterial):
    r"""A strain-based user-defined material definition with a given function for the
    stress tensor and the (fourth-order) elasticity tensor.

    Parameters
    ----------
    material : callable
        The material model formulation. Function signature must be
        ``lambda dε, εn, σn, ζn, **kwargs: dσdε, σ, ζ``. Input arguments are the strain
        increment, old strain, old stress, the list of old state variables and optional
        keyword arguments. The function must return the algorithmic consistent
        elasticity tensor, new stress and the list of new state variables. The provided
        strain and required stress quantities are selected by the framework argument.
    dim : int, optional
        The dimension of the material formulation. Default is 3.
    statevars : tuple of int, optional
        A tuple containing the shape of each state variable. Default is (0, ).
    framework : str, optional
        The framework to be used for the stress and strain formulations. "small-strain",
        "total-lagrange" and "co-rotational" are supported. Default is "small-strain".
    symmetry : bool, optional
        Take the symmetric part of the returned stress and the minor symmetric-parts of
        the algorithmic consistent elasticity tensor. Default is True. May enhance
        performance if the material returns symmetric tensors.

    Notes
    -----
    The (default) small-strain framework evaluates the strain tensor as the symmetric
    part of the displacement gradient, see Eq. :eq:`small-strain`.

    ..  math::
        :label: small-strain

        \boldsymbol{\varepsilon} = \operatorname{sym} \left(
            \frac{\partial \boldsymbol{u}}{\partial \boldsymbol{x}} \right)

    The Total-Lagrange framework uses the Green-Lagrange strain, see Eq.
    :eq:`gl-strain`,

    ..  math::
        :label: gl-strain

        \boldsymbol{E} = \frac{1}{2} \left( \boldsymbol{C} - \boldsymbol{1} \right)

    with the right Cauchy-Green deformation tensor, as denoted in Eq.
    :eq:`right-cauchy-green-deformation`.

    ..  math::
        :label: right-cauchy-green-deformation

        \boldsymbol{C} = \boldsymbol{F}^T \boldsymbol{F}

    Within the Total-Lagrange framework, the second Piola-Kirchhoff stress tensor
    :math:`\boldsymbol{S}`, as a function of the Green-Lagrange strain tensor
    :math:`\boldsymbol{E}`, is converted to the first Piola-Kirchhoff stress tensor
    :math:`\boldsymbol{P}`, see Eq. :eq:`pk1-stress`.

    ..  math::
        :label: pk1-stress

        \boldsymbol{P} = \boldsymbol{F}\ \boldsymbol{S}

    Furthermore, the fourth-order material elasticity tensor in the Total-Lagrange
    framework is also converted for algorithmic consistency, see Eq.
    :eq:`pk1-elasticity`.

    ..  math::
        :label: pk1-elasticity

        \mathbb{A}_{iJkL} = F_{iI}\ F_{kK}\ \mathbb{C}_{IJKL}
            + \delta_{ik}\ S_{JL}

    ..  warning::

        The implementation of the co-rotational framework is experimental. The fourth-
        order elasticity tensor is an approximation and hence, Newton's method does not
        converge quadratically.

    Examples
    --------
    Take this code-block for a linear-elastic material formulation

    ..  plot::
        :context: close-figs

        import felupe as fem
        from felupe.math import identity, cdya_ik, dya, trace

        def linear_elastic(dε, εn, σn, ζn, λ, μ, **kwargs):
            '''3D linear-elastic material formulation.

            Arguments
            ---------
            dε : ndarray
                Incremental strain tensor.
            εn : ndarray
                Old strain tensor.
            σn : ndarray
                Old stress tensor.
            ζn : ndarray
                Old state variables.
            λ : float
                First Lamé-constant.
            μ : float
                Second Lamé-constant (shear modulus).
            '''

            # change of stress due to change of strain
            I = identity(dε)
            dσ = 2 * μ * dε + λ * trace(dε) * I

            # update stress and evaluate elasticity tensor
            σ = σn + dσ
            dσdε = 2 * μ * cdya_ik(I, I) + λ * dya(I, I)

            # update state variables (not used here)
            ζ = ζn

            return dσdε, σ, ζ

        umat = fem.MaterialStrain(material=linear_elastic, μ=1, λ=2)
        ax = umat.plot()

    or this minimal header as template.

    ..  plot::
        :context: close-figs

        def fun(dε, εn, σn, ζn, **kwargs):
            return dσdε, σ, ζ

        umat = fem.MaterialStrain(material=fun, **kwargs)

    The Total-Lagrange framework changes the linear-elastic material formulation to the
    Saint-Venant Kirchhoff material model formulation.

    ..  plot::
        :context: close-figs

        umat = fem.MaterialStrain(
            material=fem.linear_elastic, μ=1, λ=2, framework="total-lagrange"
        )
        ax = umat.plot(
            ux=fem.math.linsteps([1, 1.5], num=20),
            ps=fem.math.linsteps([1, 1.5], num=20),
            bx=fem.math.linsteps([1, 1.25], num=10),
        )

    See Also
    --------
    linear_elastic : 3D linear-elastic material formulation
    linear_elastic_plastic_isotropic_hardening : Linear-elastic-plastic material
        formulation with linear isotropic hardening (return mapping algorithm).
    LinearElasticPlasticIsotropicHardening : Linear-elastic-plastic material formulation
        with linear isotropic hardening (return mapping algorithm).

    """

    def __init__(
        self,
        material,
        dim=3,
        statevars=(0,),
        framework="small-strain",
        symmetry=True,
        **kwargs,
    ):
        self.material = material
        self.statevars_shape = statevars
        self.statevars_size = [np.prod(shape) for shape in statevars]
        self.statevars_offsets = np.cumsum(self.statevars_size)
        self.nstatevars = sum(self.statevars_size)

        self.framework = framework
        self.symmetry = symmetry

        self.kwargs = {**kwargs, "tangent": None}

        self.dim = dim
        self.x = [np.eye(dim), np.zeros(2 * dim**2 + self.nstatevars)]

        self.stress = self.gradient
        self.elasticity = self.hessian

    def extract(self, x):
        "Extract the input and evaluate strains, stresses and state variables."

        # unpack deformation gradient F = dx/dX
        dim = self.dim
        dxdX, statevars = x

        rotation = None

        if self.framework == "small-strain":
            strain = sym(dxdX - identity(dxdX))

        elif self.framework == "total-lagrange":
            strain = (dot(transpose(dxdX), dxdX) - identity(dxdX)) / 2

        elif self.framework == "co-rotational":
            svd_res = svd(dxdX)
            rotation = dot(svd_res.U, svd_res.Vh)
            stretch = dot(transpose(rotation), dxdX)
            strain = stretch - identity(dxdX)

        else:
            raise NotImplementedError(f'Framework "{self.framework}" not implemented.')

        # separate strain and stress from state variables
        statevars_all = np.split(
            statevars, [*self.statevars_offsets, self.nstatevars + dim**2]
        )
        strain_old_1d, stress_old_1d = statevars_all[-2:]

        # list of state variables with original shapes
        shapes = self.statevars_shape
        statevars_old = [
            reshape(sv, shape).copy() for sv, shape in zip(statevars_all[:-2], shapes)
        ]

        # reshape strain and stress from (dim**2,) to (dim, dim)
        strain_old = strain_old_1d.reshape(dim, dim, *strain_old_1d.shape[1:])
        stress_old = stress_old_1d.reshape(dim, dim, *stress_old_1d.shape[1:])

        # change of strain
        dstrain = strain - strain_old

        return strain_old, dstrain, stress_old, statevars_old, rotation

    def gradient(self, x):
        strain_old, dstrain, stress_old, statevars_old, rotation = self.extract(x)
        self.kwargs["tangent"] = False

        # unpack deformation gradient F = dx/dX
        dxdX, statevars = x

        dsde, stress_new, statevars_new_list = self.material(
            dstrain, strain_old, stress_old, statevars_old, **self.kwargs
        )

        strain_new_1d = (strain_old + dstrain).reshape(-1, *strain_old.shape[2:])
        stress_new_1d = stress_new.reshape(-1, *strain_old.shape[2:])

        statevars_new = np.concatenate(
            [
                *[ravel(sv) for sv in statevars_new_list],
                strain_new_1d,
                stress_new_1d,
            ],
            axis=0,
        )

        if self.symmetry:
            stress_new = sym(stress_new)

        if self.framework == "total-lagrange":
            # convert second to first Piola-Kirchhoff stress
            stress_new = dot(dxdX, stress_new)

        if self.framework == "co-rotational":
            stress_new = dot(rotation, stress_new)

        return [stress_new, statevars_new]

    def hessian(self, x):
        strain_old, dstrain, stress_old, statevars_old, rotation = self.extract(x)
        self.kwargs["tangent"] = True

        # unpack deformation gradient F = dx/dX
        dxdX, statevars = x

        dsde, stress_new, statevars_new_list = self.material(
            dstrain, strain_old, stress_old, statevars_old, **self.kwargs
        )
        dsde = np.ascontiguousarray(dsde)

        if self.symmetry:

            # enforce minor symmetries on the algorithmic consistent elasticity tensor
            dsde = np.add(dsde, np.einsum("ijlk...->ijkl...", dsde), out=dsde)
            dsde = np.add(dsde, np.einsum("jikl...->ijkl...", dsde), out=dsde)
            dsde = np.multiply(dsde, 0.25, out=dsde)

        if self.framework in ["total-lagrange", "co-rotational"]:

            phi = {
                "total-lagrange": dxdX,
                "co-rotational": rotation,
            }[self.framework]

            # convert elasticity tensor
            dsde = np.einsum("iI...,kK...,IJKL...->iJkL...", phi, phi, dsde)

            if self.symmetry:
                stress_new = sym(stress_new)

            geometric = cdya_ik(np.eye(3), stress_new)
            dsde = np.sum(np.broadcast_arrays(dsde, geometric), axis=0)

        return [dsde]



from ._blatz_ko import blatz_ko
from ._extended_tube import extended_tube
from ._miehe_goektepe_lulei import miehe_goektepe_lulei
from ._mooney_rivlin import mooney_rivlin
from ._neo_hooke import neo_hooke
from ._ogden import ogden
from ._storakers import storakers
from ._third_order_deformation import third_order_deformation
from ._van_der_waals import van_der_waals
from ._yeoh import yeoh

__all__ = [
    "blatz_ko",
    "extended_tube",
    "miehe_goektepe_lulei",
    "mooney_rivlin",
    "neo_hooke",
    "ogden",
    "storakers",
    "third_order_deformation",
    "van_der_waals",
    "yeoh",
]

# default (stable) material parameters
blatz_ko.kwargs = dict(mu=0)
extended_tube.kwargs = dict(Gc=0, Ge=0, beta=1, delta=0)
miehe_goektepe_lulei.kwargs = dict(mu=0, N=100, U=0, p=2, q=2)
mooney_rivlin.kwargs = dict(C10=0, C01=0)
neo_hooke.kwargs = dict(mu=0)
ogden.kwargs = dict(mu=[0, 0], alpha=[2, -2])
storakers.kwargs = dict(mu=[0], alpha=[2], beta=[1])
third_order_deformation.kwargs = dict(C10=0, C01=0, C11=0, C20=0, C30=0)
van_der_waals.kwargs = dict(mu=0, beta=0, a=0, limit=100)
yeoh.kwargs = dict(C10=0, C20=0, C30=0)



from ._linear_elastic import linear_elastic
from ._linear_elastic_plastic_isotropic import (
    LinearElasticPlasticIsotropicHardening,
    linear_elastic_plastic_isotropic_hardening,
)
from ._linear_elastic_viscoelastic import linear_elastic_viscoelastic

__all__ = [
    "linear_elastic",
    "linear_elastic_viscoelastic",
    "linear_elastic_plastic_isotropic_hardening",
    "LinearElasticPlasticIsotropicHardening",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from numpy import array
from tensortrax.math import base, einsum, log
from tensortrax.math import sum as tsum
from tensortrax.math.linalg import eigh
from tensortrax.math.special import from_triu_1d

from .....math import cross


def saint_venant_kirchhoff_orthotropic(C, mu, lmbda, r1, r2, r3=None, k=2):
    r"""Strain energy function of the orthotropic hyperelastic
    `Saint-Venant Kirchhoff <https://en.wikipedia.org/wiki/Hyperelastic_material#Saint_Venant-Kirchhoff_model>`_
    material formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    mu : list of float
        List of the three second Lamé parameters :math:`\mu_1,\mu_2, \mu_3`.
    lmbda : list of float
        List of six (upper triangle) first Lamé parameters :math:`\lambda_{11},
        \lambda_{12}, \lambda_{13}, \lambda_{22}, \lambda_{23}, \lambda_{33}`.
    r1 : list of float
        First normal vector of planes of symmetry.
    r2 : list of float
        Second normal vector of planes of symmetry.
    r3 : list of float or None, optional
        Third normal vector of planes of symmetry. If None, the third normal vector
        is evaluated as :math:`r_1 \times r_2`. Default is None.
    k : float, optional
        Strain exponent (default is 2). If 2, the Green-Lagrange strain measure is used.
        For any other value, the family of Seth-Hill strains is used.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-svk-ortho`

    ..  math::
        :label: psi-svk-ortho

        \psi = \sum_{a=1}^3
            \mu_a \boldsymbol{E} : \boldsymbol{r}_a \otimes \boldsymbol{r}_a +
            \sum_{a=1}^3 \sum_{b=1}^3 \frac{\lambda_{ab}}{2}
                \left(\boldsymbol{E} : \boldsymbol{r}_a \otimes \boldsymbol{r}_a \right)
                \left(\boldsymbol{E} : \boldsymbol{r}_b \otimes \boldsymbol{r}_b \right)

    Examples
    --------
    ..  warning::
        The orthotropic Saint-Venant Kirchhoff material formulation is unstable for
        large strains.

    ..  plot::

        >>> import felupe.constitution.tensortrax as mat
        >>> import felupe as fem
        >>>
        >>> r = fem.math.rotation_matrix(0, axis=2)
        >>> lmbda, mu = fem.constitution.lame_converter_orthotropic(
        ...     E=[10, 10, 10],
        ...     nu=[0.3, 0.3, 0.3],
        ...     G=[1, 1, 1],
        ... )
        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.saint_venant_kirchhoff_orthotropic,
        ...     mu=mu,
        ...     lmbda=lmbda,
        ...     r1=r[:, 0],
        ...     r2=r[:, 1],
        ...     r3=r[:, 2],
        ... )
        >>> ax = umat.plot(ux=fem.math.linsteps([1, 1.1], num=10), ps=None, bx=None)

    """
    eye = base.eye

    if k == 2:
        E = (C - eye(C)) / 2

    else:
        λ2, M = eigh(C)
        if k == 0:
            E = tsum(log(λ2) / 2 * M, axis=0)
        else:
            E = tsum((λ2 ** (k / 2) - 1) / k * M, axis=0)

    μ = array(mu)
    λ = from_triu_1d(array(lmbda))

    if r3 is None:
        r3 = cross(r1, r2)

    r = array([r1, r2, r3])
    Err = einsum("ai...,ij...,aj...->a...", r, E, r)

    λI1 = einsum("ab,a...,b...->...", λ / 2, Err, Err)
    μI2 = einsum("a,ai...,ij...,aj...->...", μ, r, E @ E, r)

    return μI2 + λI1



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from .microsphere import langevin, linear, nonaffine_stretch, nonaffine_tube


def miehe_goektepe_lulei(C, mu, N, U, p, q):
    """Strain energy function of the isotropic hyperelastic
    `micro-sphere <https://doi.org/10.1016/j.jmps.2004.03.011>`_ model formulation [1]_.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    mu : float
        Shear modulus (ground state stifness).
    N : float
        Number of chain segments (chain locking response).
    U : float
        Tube geometry parameter (3D locking characteristics).
    p : float
        Non-affine stretch parameter (additional constraint stifness).
    q : float
        Non-affine tube parameter (shape of constraint stress).

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> import felupe as fem
        >>>
        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.miehe_goektepe_lulei,
        ...     mu=0.1475,
        ...     N=3.273,
        ...     p=9.31,
        ...     U=9.94,
        ...     q=0.567,
        ... )
        >>> ux = ps = fem.math.linsteps([1, 2], num=50)
        >>> bx = fem.math.linsteps([1, 1.5], num=50)
        >>> ax = umat.plot(ux=ux, ps=ps, bx=bx, incompressible=True)

    References
    ----------
    .. [1] C. Miehe, S. Göktepe and F. Lulei, "A micro-macro approach to rubber-like
       materials - Part I: the non-affine micro-sphere model of rubber elasticity",
       Journal of the Mechanics and Physics of Solids, vol. 52, no. 11. Elsevier BV, pp.
       2617–2660, Nov. 2004. doi:
       `10.1016/j.jmps.2004.03.011 <https://doi.org/10.1016/j.jmps.2004.03.011>`_.
    """

    kwargs_stretch = {"mu": mu, "N": N}
    kwargs_tube = {"mu": mu * N * U}

    return nonaffine_stretch(
        C, p=p, f=langevin, kwargs=kwargs_stretch
    ) + nonaffine_tube(C, q=q, f=linear, kwargs=kwargs_tube)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ...math import ddot, identity, transpose
from .._base import ConstitutiveMaterial


class LinearElasticOrthotropic(ConstitutiveMaterial):
    r"""Orthotropic linear-elastic material formulation.

    Parameters
    ----------
    E : float
        Young's modulus (E1, E2, E3).
    nu : float
        Poisson ratio (nu12, nu23, n31).
    G : float
        Shear modulus (G12, G23, G31).

    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.LinearElasticOrthotropic(
        ...     E=[1, 1, 1], nu=[0.3, 0.3, 0.3], G=[0.4, 0.4, 0.4]
        ... )
        >>> ax = umat.plot()

    """

    def __init__(self, E, nu, G):
        self.E = E
        self.nu = nu
        self.G = G

        self.kwargs = {"E": self.E, "nu": self.nu, "G": self.G}

        # aliases for gradient and hessian
        self.stress = self.gradient
        self.elasticity = self.hessian

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [np.eye(3), np.zeros(0)]

    def gradient(self, x):
        r"""Evaluate the stress tensor (as a function of the deformation gradient).

        Parameters
        ----------
        x : list of ndarray
            List with Deformation gradient :math:`\boldsymbol{F}` (3x3) as first item.

        Returns
        -------
        ndarray of shape (3, 3, ...)
            Stress tensor

        """

        F, statevars = x[0], x[-1]

        # convert the deformation gradient to strain
        H = F - identity(F)
        strain = (H + transpose(H)) / 2

        # init stress
        elast = self.hessian(x=x)[0]

        return [ddot(elast, strain, mode=(4, 2)), statevars]

    def hessian(self, x=None, shape=(1, 1), dtype=None):
        r"""Evaluate the elasticity tensor. The Deformation gradient is only
        used for the shape of the trailing axes.

        Parameters
        ----------
        x : list of ndarray, optional
            List with Deformation gradient :math:`\boldsymbol{F}` (3x3) as first item
            (default is None).
        shape : tuple of int, optional
            Tuple with shape of the trailing axes (default is (1, 1)).

        Returns
        -------
        ndarray of shape (3, 3, 3, 3, ...)
            elasticity tensor

        """

        E1, E2, E3 = self.E
        nu12, nu23, nu31 = self.nu
        G12, G23, G31 = self.G

        if x is not None:
            dtype = x[0].dtype

        elast = np.zeros((3, 3, 3, 3, *shape), dtype=dtype)

        nu21 = nu12 * E2 / E1
        nu32 = nu23 * E3 / E2
        nu13 = nu31 * E1 / E3

        J = 1 / (1 - nu12 * nu21 - nu23 * nu32 - nu31 * nu13 - 2 * nu21 * nu32 * nu13)

        elast[0, 0, 0, 0] = E1 * (1 - nu23 * nu32) * J
        elast[1, 1, 1, 1] = E2 * (1 - nu13 * nu31) * J
        elast[2, 2, 2, 2] = E3 * (1 - nu12 * nu21) * J

        elast[0, 0, 1, 1] = elast[1, 1, 0, 0] = E1 * (nu21 + nu31 * nu23) * J
        elast[0, 0, 2, 2] = elast[2, 2, 0, 0] = E1 * (nu31 + nu21 * nu32) * J
        elast[1, 1, 2, 2] = elast[2, 2, 1, 1] = E2 * (nu32 + nu12 * nu31) * J

        elast[
            [0, 1, 0, 1],
            [1, 0, 1, 0],
            [0, 0, 1, 1],
            [1, 1, 0, 0],
        ] = G12

        elast[
            [0, 2, 0, 2],
            [2, 0, 2, 0],
            [0, 0, 2, 2],
            [2, 2, 0, 0],
        ] = G31

        elast[
            [1, 2, 1, 2],
            [2, 1, 2, 1],
            [1, 1, 2, 2],
            [2, 2, 1, 1],
        ] = G23

        return [elast]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ._scene import Scene


class ViewMesh(Scene):
    r"""Provide Visualization methods for :class:`felupe.Mesh` with optional given
    dicts of point- and cell-data items.

    Parameters
    ----------
    mesh : felupe.Mesh
        The mesh object.
    point_data : dict or None, optional
        Additional point-data dict (default is None).
    cell_data : dict or None, optional
        Additional cell-data dict (default is None).
    cell_type : pyvista.CellType or None, optional
        Cell-type of PyVista (default is None).

    Attributes
    ----------
    mesh : pyvista.UnstructuredGrid
        A generalized Dataset with the mesh as well as point- and cell-data. This is
        not an instance of :class:`felupe.Mesh`.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=3)
        >>> displacement = np.arange(81).reshape(27, 3) / 300
        >>> view = fem.ViewMesh(mesh, point_data={"Displacement": displacement})
        >>>
        >>> view.plot("Displacement", component=None).show()

    See Also
    --------
    felupe.view.Scene : Base class for plotting a static scene.
    felupe.ViewField : Provide Visualization methods for a field container.
    felupe.ViewSolid : Provide Visualization methods for a field container or a
        solid body.

    """

    def __init__(self, mesh, point_data=None, cell_data=None, cell_type=None):
        self.mesh = mesh.as_pyvista(cell_type=cell_type)

        if point_data is None:
            point_data = {}

        if cell_data is None:
            cell_data = {}

        for label, data in point_data.items():
            self.mesh.point_data[label] = data

        for label, data in cell_data.items():
            self.mesh.cell_data[label] = data

        self.mesh.set_active_scalars(None)
        self.mesh.set_active_vectors(None)
        self.mesh.set_active_tensors(None)



from jax.numpy import einsum, sqrt
from jax.numpy.linalg import det, inv

from ......quadrature import BazantOh


def nonaffine_stretch(C, p, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Non-affine stretch part."

    r = quadrature.points
    w = quadrature.weights

    # affine stretches
    λ = sqrt(einsum("ai,ij...,aj->a...", r, C, r))

    # non-affine stretch (distortional part)
    Λ = det(C) ** (-1 / 6) * einsum("a...,a->...", λ**p, w) ** (1 / p)

    return f(Λ, **kwargs)


def nonaffine_tube(C, q, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Non-affine tube part."

    r = quadrature.points
    w = quadrature.weights

    # affine area-stretches
    λa = sqrt(einsum("ai,ij...,aj->a...", r, inv(C), r))

    # non-affine tube contraction (distortional part)
    Λt = det(C) ** (q / 6) * einsum("a...,a->...", λa**q, w)

    return f(Λt, **kwargs)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import trace
from tensortrax.math.linalg import det


def yeoh(C, C10, C20, C30):
    r"""Strain energy function of the isotropic hyperelastic
    `Yeoh <https://en.wikipedia.org/wiki/Yeoh_(hyperelastic_model)>`_
    material formulation.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    C10 : float
        Material parameter associated to the linear term of the first invariant.
    C20 : float
        Material parameter associated to the quadratic term of the first invariant.
    C30 : float
        Material parameter associated to the cubic term of the first invariant.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-yeoh`

    ..  math::
        :label: psi-yeoh

        \psi = C_{10} \left(\hat{I}_1 - 3 \right) + C_{20} \left(\hat{I}_1 - 3 \right)^2
             + C_{30} \left(\hat{I}_1 - 3 \right)^3

    with the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor, see Eq. :eq:`invariants-yeoh`.

    ..  math::
        :label: invariants-yeoh

        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

    The :math:`C_{10}` material parameter is equal to half the initial shear modulus
    :math:`\mu` as denoted in Eq. :eq:`shear-modulus-yeoh`.

    ..  math::
        :label: shear-modulus-yeoh

        \mu = 2 C_{10}

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the hyperelastic material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.yeoh, C10=0.5, C20=-0.1, C30=0.02
        ... )
        >>> ax = umat.plot(incompressible=True)

    """

    I1 = det(C) ** (-1 / 3) * trace(C)
    return C10 * (I1 - 3) + C20 * (I1 - 3) ** 2 + C30 * (I1 - 3) ** 3



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import inspect
import warnings

import numpy as np

from ..assembly import IntegralForm
from ..math import det, dot, transpose
from ..view import ViewSolid
from ._helpers import Assemble, Evaluate, Results


class Solid:
    "Base class for solid bodies which provides methods for visualisations."

    def view(
        self,
        point_data=None,
        cell_data=None,
        cell_type=None,
        project=None,
        stress_type="Cauchy",
    ):
        """View the solid with optional given dicts of point- and cell-data items.

        Parameters
        ----------
        point_data : dict or None, optional
            Additional point-data dict (default is None).
        cell_data : dict or None, optional
            Additional cell-data dict (default is None).
        cell_type : pyvista.CellType or None, optional
            Cell-type of PyVista (default is None).
        project : callable or None, optional
            Project stress at quadrature-points to mesh-points (default is None).
        stress_type : str or None, optional
            The type of stress which is exported, either "Cauchy", "Kirchhoff" or None.
            If None, the first Piola-Kirchhoff stress (engineering stress in linear
            elasticity) is used. Default is "Cauchy".

        Returns
        -------
        felupe.ViewSolid
            An object which provides visualization methods for solid bodies,
            e.g. :class:`felupe.SolidBody`.

        See Also
        --------
        felupe.ViewSolid : Visualization methods for :class:`felupe.SolidBody`.
        felupe.project: Project given values at quadrature-points to mesh-points.
        felupe.topoints: Shift given values at quadrature-points to mesh-points.
        """

        return ViewSolid(
            self.field,
            solid=self,
            point_data=point_data,
            cell_data=cell_data,
            cell_type=cell_type,
            project=project,
            stress_type=stress_type,
        )

    def plot(self, *args, project=None, **kwargs):
        """Plot the solid body.

        See Also
        --------
        felupe.Scene.plot: Plot method of a scene.
        felupe.project: Project given values at quadrature-points to mesh-points.
        felupe.topoints: Shift given values at quadrature-points to mesh-points.
        """

        stress_type = "Cauchy"

        if len(args) > 0:
            name = kwargs.pop("name", args[0])

            if "Stress" in name:
                stress_type = (
                    name.lower()
                    .split("principal values of ")[-1]
                    .split("equivalent of ")[-1]
                    .split("stress")[0]
                    .rstrip()
                )

        if len(stress_type) == 0:
            stress_type = None

        return self.view(project=project, stress_type=stress_type).plot(*args, **kwargs)

    def screenshot(
        self,
        *args,
        filename="solidbody.png",
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the solid body.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        return self.plot(*args, off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )

    def imshow(self, *args, ax=None, dpi=None, **kwargs):
        """Take a screenshot of the solid body, show the image data in a figure and
        return the ax.
        """

        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(dpi=dpi)

        ax.imshow(self.screenshot(*args, filename=None, **kwargs))
        ax.set_axis_off()

        return ax


class SolidBody(Solid):
    r"""A SolidBody with methods for the assembly of sparse vectors/matrices.

    Parameters
    ----------
    umat : class
        A class which provides methods for evaluating the gradient and the hessian of
        the strain energy density function per unit undeformed volume. The function
        signatures must be ``dψdF, ζ_new = umat.gradient([F, ζ])`` for
        the gradient and ``d2ψdFdF = umat.hessian([F, ζ])`` for the hessian of
        the strain energy density function :math:`\psi(\boldsymbol{F})`, where
        :math:`\boldsymbol{F}` is the deformation gradient tensor and :math:`\zeta`
        holds the array of internal state variables.
    field : FieldContainer
        A field container with one or more fields.
    statevars : ndarray or None, optional
        Array of initial internal state variables (default is None).
    density : float or None, optional
        The density of the solid body.
    block : bool, optional
        Assemble a sparse matrix from sparse sub-blocks or assemble a sparse vector by
        stacking sparse matrices vertically (row wise). Default is True.
    apply : callable or None, optional
        Apply a callable on the assembled vectors and sparse matrices. Default is None.
    multiplier : float or None, optional
        A scale factor for the assembled vector and matrix. Default is None.

    Notes
    -----
    The total potential energy of internal forces is given in Eq. :eq:`solidbody`

    ..  math::
        :label: solidbody

        \Pi_{int}(\boldsymbol{F}) = \int_V \psi(\boldsymbol{F})\ dV

    with its variation, see Eq. :eq:`solidbody-variation`

    ..  math::
        :label: solidbody-variation

        \delta_\boldsymbol{u}(\Pi_{int}) =
            \int_V \frac{\partial \psi}{\partial \boldsymbol{F}} \ dV
        \longrightarrow \boldsymbol{f}_\boldsymbol{u}

    and linearization, see Eq. :eq:`solidbody-linearization`. The right-arrows in
    Eq. :eq:`solidbody-variation` and Eq. :eq:`solidbody-linearization`
    represent the assembly into system scalars, vectors or matrices.

    ..  math::
        :label: solidbody-linearization

        \delta_\boldsymbol{u}\Delta_\boldsymbol{u}(\Pi_{int}) =
            \int_V \delta\boldsymbol{F} :
                \frac{\partial^2 \psi}{
                    \partial \boldsymbol{F}\ \partial \boldsymbol{F}
                } : \Delta\boldsymbol{F}\ dV
        \longrightarrow \boldsymbol{K}_{\boldsymbol{u}\boldsymbol{u}}

    The displacement-based formulation leads to a linearized equation system as given in
    Eq. :eq:`solidbody-final`.

    ..  math::
        :label: solidbody-final

        \boldsymbol{K}_{\boldsymbol{u}\boldsymbol{u}} \cdot \delta \boldsymbol{u} +
            \boldsymbol{f}_\boldsymbol{u} = \boldsymbol{0}

    ..  note::
        This class also supports ``umat`` with mixed-field formulations like
        :class:`~felupe.NearlyIncompressible` or :class:`~felupe.ThreeFieldVariation`.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=6)
        >>> region = fem.RegionHexahedron(mesh)
        >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
        >>> boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
        >>>
        >>> umat = fem.NeoHooke(mu=1, bulk=2)
        >>> solid = fem.SolidBody(umat, field)
        >>>
        >>> table = fem.math.linsteps([0, 1], num=5)
        >>> step = fem.Step(
        ...     items=[solid],
        ...     ramp={boundaries["move"]: table},
        ...     boundaries=boundaries,
        ... )
        >>>
        >>> job = fem.Job(steps=[step]).evaluate()
        >>> solid.plot("Principal Values of Cauchy Stress").show()

    See Also
    --------
    felupe.SolidBodyNearlyIncompressible : A (nearly) incompressible solid body with
        methods for the assembly of sparse vectors/matrices.
    """

    def __init__(
        self,
        umat,
        field,
        statevars=None,
        density=None,
        block=True,
        apply=None,
        multiplier=None,
    ):
        self.umat = umat
        self.field = field
        self.density = density
        self.block = block
        self.apply = apply

        self.results = Results(stress=True, elasticity=True)
        self.results.kinematics = self._extract(self.field)

        if statevars is not None:
            # state variables are located in the results
            self.results.statevars = statevars

        else:
            # init the shape of the state variables
            statevars_shape = (0,)

            # take the shape of the last item of the list if the provided umat has an
            # x-attribute
            if hasattr(umat, "x"):
                statevars_shape = umat.x[-1].shape

            # create and fill the array of state variables with zeros
            # state variables are located in the results
            self.results.statevars = np.zeros(
                (
                    *statevars_shape,
                    field.region.quadrature.npoints,
                    field.region.mesh.ncells,
                )
            )

        self.assemble = Assemble(
            vector=self._vector,
            matrix=self._matrix,
            mass=self._mass,
            multiplier=multiplier,
        )

        self.evaluate = Evaluate(
            gradient=self._gradient,
            hessian=self._hessian,
            cauchy_stress=self._cauchy_stress,
            kirchhoff_stress=self._kirchhoff_stress,
        )

    def checkpoint(self):
        """Return a checkpoint of the solid body.

        Returns
        -------
        dict
            A dict with checkpoint arrays / objects.

        Examples
        --------
        ..  pyvista-plot::
            :force_static:

            import felupe as fem

            mesh = fem.Rectangle(b=(3, 1), n=(7, 3))
            region = fem.RegionQuad(mesh)
            field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])

            umat = fem.NeoHooke(mu=1, bulk=2)
            solid = fem.SolidBody(umat=umat, field=field)

            # 1. vertical compression
            boundaries = fem.dof.uniaxial(
                field, clamped=True, sym=False, axis=1, return_loadcase=False
            )
            ramp = {boundaries["move"]: fem.math.linsteps([0, -0.2], num=2)}
            step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
            job = fem.Job(steps=[step]).evaluate()

            # checkpoint the current state of deformation
            checkpoint = solid.checkpoint()

            # 2a. horizontal shear (right)
            boundaries = fem.dof.shear(
                field, sym=False, moves=(0, 0, -0.2), return_loadcase=False
            )
            ramp = {boundaries["move"]: fem.math.linsteps([0, 1], num=2)}
            step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
            job = fem.Job(steps=[step]).evaluate()

            # (1.) restore compression without shear
            solid.restore(checkpoint)

            # 2b. horizontal shear (left)
            boundaries = fem.dof.shear(
                field, sym=False, moves=(0, 0, -0.2), return_loadcase=False
            )
            ramp = {boundaries["move"]: fem.math.linsteps([0, -1], num=2)}
            step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
            job = fem.Job(steps=[step]).evaluate()

        See Also
        --------
        felupe.SolidBody.restore : Restore a checkpoint of a solid body inplace.
        """

        return {
            **self.field.checkpoint(),
            "results.statevars": self.results.statevars.copy(),
        }

    def restore(self, checkpoint, restore_statevars=True):
        """Restore a checkpoint inplace.

        Parameters
        ----------
        checkpoint : dict
            A dict with checkpoint arrays / objects.
        restore_statevars : bool, optional
            Flag to restore state variables. This is a power feature and must be used
            with caution! Default is True.

        See Also
        --------
        felupe.SolidBody.checkpoint : Return a checkpoint of the solid body.
        """

        self.field.restore(checkpoint)
        if restore_statevars:
            self.results.statevars[:] = checkpoint["results.statevars"]

        # results must be re-evaluated
        self.evaluate.gradient(self.field)
        self.evaluate.hessian(self.field)

        # reset force and stiffness
        self.results.force = None
        self.results.stiffness = None

    def revolve(self, n=11, phi=180):
        """Return a revolved solid body.

        Parameters
        ----------
        n : int, optional
            Number of n-point revolutions (or (n-1) cell revolutions), default is 11.
        phi : float or ndarray, optional
            Revolution angle in degree (default is 180).

        Returns
        -------
        SolidBody
            The revolved solid body.

        Examples
        --------
        First, create an axisymmetric model. A non-homogeneous uniaxial tension load
        case is applied on the solid body.

        ..  pyvista-plot::
            :context:
            :force_static:

            >>> import felupe as fem
            >>>
            >>> region = fem.RegionQuad(mesh=fem.Rectangle(n=6))
            >>> field = fem.FieldContainer([fem.FieldAxisymmetric(region, dim=2)])
            >>>
            >>> umat = fem.NeoHookeCompressible(mu=1, lmbda=2)
            >>> solid = fem.SolidBody(umat=umat, field=field)
            >>>
            >>> boundaries = fem.dof.uniaxial(
            ...     solid.field, clamped=True, sym=False, return_loadcase=False
            ... )
            >>> step = fem.Step(items=[solid], boundaries=boundaries)
            >>> job = fem.Job(steps=[step]).evaluate()
            >>>
            >>> solid.plot("Principal Values of Cauchy Stress").show()

        The solid body is now revolved around the x-axis. This model may now be used to
        apply non-axisymmetric loads. Here, the same load case is also applied on the
        3d-model.

        ..  pyvista-plot::
            :context:
            :force_static:

            >>> new_solid = solid.revolve(n=11, phi=180)
            >>> boundaries = fem.dof.uniaxial(
            ...     new_solid.field, clamped=True, sym=(0, 0, 1), return_loadcase=False
            ... )
            >>>
            >>> step = fem.Step(items=[new_solid], boundaries=boundaries)
            >>> job = fem.Job(steps=[step]).evaluate()
            >>>
            >>> new_solid.plot("Principal Values of Cauchy Stress").show()

        See Also
        --------
        SolidBodyNearlyIncompressible.revolve : Return a revolved solid body

        """

        new_field = self.field.revolve(n=n, phi=phi)

        # create a new solid body
        new_solid = SolidBody(
            umat=self.umat,
            field=new_field,
            density=self.density,
            block=self.block,
            apply=self.apply,
            multiplier=self.assemble.multiplier,
        )

        # expand state variables (no rotation is applied here)
        new_shape = np.array(new_solid.results.statevars.shape)
        old_shape = np.array(self.results.statevars.shape)

        reps = np.zeros(len(new_shape), dtype=int)
        reps[0] = new_shape[0]
        reps[1:] = new_shape[1:] // old_shape[1:]

        new_solid.results.statevars[:] = np.tile(self.results.statevars, reps)

        return new_solid

    def _vector(
        self,
        field=None,
        parallel=False,
        items=None,
        args=(),
        kwargs=None,
        block=None,
        apply=None,
    ):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field

        if block is None:
            block = self.block

        if apply is None:
            apply = self.apply

        # evaluate the (first Piola-Kirchhoff) stress tensor and store it in the results
        self.results.stress = self._gradient(field, args=args, kwargs=kwargs)

        # assemble the internal force vector
        # optionally, use only the first n items for mixed-field formulations
        self.results.force = IntegralForm(
            fun=self.results.stress[slice(items)],
            v=self.field,
            dV=self.field.region.dV,
        ).assemble(parallel=parallel, block=block)

        # apply a callback on the assembled internal force vector
        if apply is not None:
            self.results.force = apply(self.results.force)

        return self.results.force

    def _matrix(
        self,
        field=None,
        parallel=False,
        items=None,
        args=(),
        kwargs=None,
        block=None,
        apply=None,
    ):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field

        if block is None:
            block = self.block

        if apply is None:
            apply = self.apply

        # evaluate the fourth-order elasticity tensor and store it in the results
        # (associated to the first Piola-Kirchhoff stress tensor, i.e. the partial
        # derivative of the first Piola-Kirchhoff stress tensor w.r.t. the deformation
        # gradient tensor)
        self.results.elasticity = self._hessian(field, args=args, kwargs=kwargs)

        # assemble the (sparse) tangent stiffness matrix
        # optionally, use only the first n items for mixed-field formulations
        form = IntegralForm(
            fun=self.results.elasticity[slice(items)],
            v=self.field,
            u=self.field,
            dV=self.field.region.dV,
        )

        # in a first step, integrate the weak-form and store the stiffness values
        # (this dense array is only allocated once and will be re-used in all following
        # evaluations)
        self.results.stiffness_values = form.integrate(
            parallel=parallel, out=self.results.stiffness_values
        )

        # finally, the dense array of stiffness-values is assembled into the sparse
        # matrix
        self.results.stiffness = form.assemble(
            values=self.results.stiffness_values, block=block
        )

        # apply a callback on the assembled tangent stiffness matrix
        if apply is not None:
            self.results.stiffness = apply(self.results.stiffness)

        return self.results.stiffness

    def _extract(self, field):
        "Evaluate and return the kinematics (the deformation gradient tensor)."

        self.field = field
        self.results.kinematics = self.field.extract(out=self.results.kinematics)

        return self.results.kinematics

    def _gradient(self, field=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        # update the deformation gradient
        if field is not None:
            self.field = field
            self.results.kinematics = self._extract(self.field)

        if "out" in inspect.signature(self.umat.gradient).parameters:
            kwargs["out"] = self.results.gradient

        # evaluate the partial derivative of the strain energy density w.r.t. the
        # deformation gradient
        gradient = self.umat.gradient(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )
        # store the first Piola-Kirchhoff stress tensor in the results
        self.results.gradient = gradient[0]

        # store all gradients and the temporary state variables in the results
        # the state variables are only updated to the public results-attribute if the
        # solution is valid
        self.results.stress, self.results._statevars = gradient[:-1], gradient[-1]

        return self.results.stress

    def _hessian(self, field=None, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        if field is not None:
            self.field = field
            self.results.kinematics = self._extract(self.field)

        if "out" in inspect.signature(self.umat.hessian).parameters:
            kwargs["out"] = self.results.hessian

        # evaluate the second partial derivative of the strain energy density w.r.t. the
        # deformation gradient
        self.results.elasticity = self.umat.hessian(
            [*self.results.kinematics, self.results.statevars], *args, **kwargs
        )

        # store the partial derivative of the first Piola-Kirchhoff stress tensor w.r.t.
        # the deformation gradient in the results
        self.results.hessian = self.results.elasticity[0]

        return self.results.elasticity

    def _kirchhoff_stress(self, field=None):
        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]

        return dot(P, transpose(F))

    def _cauchy_stress(self, field=None):
        self._gradient(field)

        P = self.results.stress[0]
        F = self.results.kinematics[0]

        if P.shape[:2] == (2, 2):
            warnings.warn(
                "\n".join(
                    [
                        "Cauchy stress tensor can't be evaluated on a 2d-Field.",
                        "Falling-back to the Kirchhoff stress tensor.",
                    ]
                )
            )
            J = 1
        else:
            J = det(F)

        return dot(P, transpose(F)) / J

    def _mass(self, density=None):
        if density is None:
            density = self.density

        field = self.field[0].as_container()
        dim = field[0].dim

        form = IntegralForm(
            fun=[density * np.eye(dim).reshape(dim, dim, 1, 1)],
            v=field,
            u=field,
            dV=field.region.dV,
            grad_v=[False],
            grad_u=[False],
        )

        return form.assemble()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from scipy.sparse import csr_matrix

from ._helpers import Assemble, Results


class PointLoad:
    r"""A point load with methods for the assembly of sparse vectors/matrices, applied
    on the n-th field.

    Parameters
    ----------
    field : FieldContainer
        A field container with fields created on a region.
    points : list of int
        A list with point ids where the values are applied.
    values : float or array_like or None, optional
        Values at points (default is None). If None, the values are set to zero.
    apply_on : int, optional
        The n-th field on which the point load is applied (default is 0).
    axisymmetric : bool, optional
        A flag to multiply the assembled vector and matrix by a scaling factor of
        :math:`2 \pi` (default is False).

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.mesh.Line(n=3)
        >>> element = fem.element.Line()
        >>> quadrature = fem.GaussLegendre(order=1, dim=1)
        >>>
        >>> region = fem.Region(mesh, element, quadrature)
        >>> field = fem.FieldContainer([fem.Field(region, dim=1)])
        >>>
        >>> load = fem.PointLoad(field, [1, 2], values=[[3], [5]])
        >>>
        >>> vector = load.assemble.vector()
        >>> vector.toarray()
        array([[0.],
               [3.],
               [5.]])
    """

    def __init__(self, field, points, values=None, apply_on=0, axisymmetric=False):
        self.field = field
        self.points = points

        if values is None:
            self.values = 0
        else:
            self.values = values

        self.apply_on = apply_on
        self.axisymmetric = axisymmetric

        self.results = Results()
        self.assemble = Assemble(
            vector=self._vector, matrix=self._matrix, multiplier=-1.0
        )

    def update(self, values):
        self.__init__(self.field, self.points, values, self.apply_on, self.axisymmetric)

    def _vector(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        force = [np.zeros_like(f.values) for f in self.field.fields]
        force[self.apply_on][self.points] += self.values

        if self.axisymmetric:
            mesh_points = self.field[0].region.mesh.points
            radius = mesh_points[self.points, 1].reshape(-1, 1)
            force[self.apply_on][self.points] *= 2 * np.pi * radius

        self.results.force = csr_matrix(
            np.concatenate([f.ravel() for f in force]).reshape(-1, 1)
        )

        return self.results.force

    def _matrix(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        n = np.sum(self.field.fieldsizes)
        self.results.stiffness = csr_matrix(([0.0], ([0], [0])), shape=(n, n))

        return self.results.stiffness

    def plot(
        self,
        plotter=None,
        color="red",
        scale=0.125,
        deformed=True,
        **kwargs,
    ):
        "Plot the point load."

        mesh = self.field.region.mesh

        if plotter is None:
            plotter = mesh.plot()

        if len(self.points) > 0:

            x = mesh.points

            if deformed:
                x = x + self.field[0].values

            points = np.pad(x, ((0, 0), (0, 3 - mesh.dim)))
            magnitude = max(points.max(axis=0) - points.min(axis=0)) * scale

            values = np.atleast_2d(self.values)

            skip = np.zeros(3, dtype=bool)
            skip[values.shape[1] :] = True

            if values.shape[1] > 1:
                skip[: values.shape[1]][np.isclose(values, 0).all(axis=0)] = True

            for a, (skip_axis, direction) in enumerate(zip(skip, np.eye(3))):
                d = np.broadcast_to(direction.reshape(1, 3), points[self.points].shape)
                if not skip_axis:
                    _ = plotter.add_arrows(
                        points[self.points],
                        direction=d * np.sign(values[0, a]),
                        mag=magnitude,
                        color=color,
                        **kwargs,
                    )

        return plotter



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ._scene import Scene


class ViewXdmf(Scene):  # pragma: no cover
    """Provide Visualization methods for a XDMF file generated by
    :meth:`Job.evaluate(filename="result.xdmf")`. The warped (deformed) mesh is created
    from the values of the point-data "Displacement".

    Parameters
    ----------
    filename : str
        The filename of the XDMF file (including the extension).
    time : float, optional
        The time value at which the data is extracted (default is 0).

    Attributes
    ----------
    mesh : pyvista.UnstructuredGrid
        A generalized Dataset with the mesh as well as point- and cell-data. This is
        not an instance of :class:`felupe.Mesh`.

    """

    def __init__(
        self,
        filename,
        time=0,
    ):
        "XDMF Result file reader and plotter."

        self.filename = filename
        self.mesh = self._read(time)

    def _read(self, time):
        "Read the file and obtain the mesh."

        import pyvista as pv

        self.file = pv.XdmfReader(self.filename)
        self.file.set_active_time_value(time)

        mesh = self.file.read()[0]

        return mesh

    def set_active_time_value(self, time):
        "Set new active time value and re-read the mesh."

        self.mesh = self._read(time)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import sum as tsum
from tensortrax.math.linalg import det, eigvalsh


def ogden(C, mu, alpha):
    r"""Strain energy function of the isotropic hyperelastic
    `Ogden <https://en.wikipedia.org/wiki/Ogden_(hyperelastic_model)>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor or jax.Array
        Right Cauchy-Green deformation tensor.
    mu : list of float
        List of moduli.
    alpha : list of float
        List of stretch exponents.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-ogden`

    ..  math::
        :label: psi-ogden

        \psi = \sum_i \frac{2 \mu_i}{\alpha^2_i} \left(
            \lambda_1^{\alpha_i} + \lambda_2^{\alpha_i} + \lambda_3^{\alpha_i} - 3
        \right)

    The sum of the moduli :math:`\mu_i` is equal to the initial shear modulus
    :math:`\mu`, see Eq. :eq:`shear-modulus-ogden`.

    ..  math::
        :label: shear-modulus-ogden

        \mu = \sum_i \mu_i

    Examples
    --------

    ..  plot::
        :context: close-figs

        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    ..  plot::

        >>> umat = mat.Hyperelastic(
        ...     mat.models.hyperelastic.ogden, mu=[1, 0.2], alpha=[1.7, -1.5]
        ... )
        >>> ax = umat.plot(incompressible=True)

    """

    wC = det(C) ** (-1 / 3) * eigvalsh(C)
    return tsum([2 * m / a**2 * (tsum(wC ** (a / 2)) - 3) for m, a in zip(mu, alpha)])



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ....math import cdya_ik, dya, identity, trace


def linear_elastic(dε, εn, σn, ζn, λ, μ, **kwargs):
    r"""3D linear-elastic material formulation to be used in
    :class:`~felupe.MaterialStrain`.

    Arguments
    ---------
    dε : ndarray
        Strain increment.
    εn : ndarray
        Old strain tensor.
    σn : ndarray
        Old stress tensor.
    ζn : list
        List of old state variables.
    λ : float
        First Lamé-constant.
    μ : float
        Second Lamé-constant (shear modulus).
    **kwargs
        Additional keyword arguments.

        tangent : bool
            A flag to evaluate the elasticity tensor.

    Returns
    -------
    dσdε : ndarray
        Elasticity tensor.
    σ : ndarray
        (New) stress tensor.
    ζ : list
        List of new state variables.

    Notes
    -----

    1.  Given state in point :math:`\boldsymbol{x} (\boldsymbol{\sigma}_n)` (valid).

    2.  Given strain increment :math:`\Delta\boldsymbol{\varepsilon}`, so that
        :math:`\boldsymbol{\varepsilon} = \boldsymbol{\varepsilon}_n + \Delta\boldsymbol{\varepsilon}`.

    3.  Evaluation of the stress :math:`\boldsymbol{\sigma}` and the algorithmic
        consistent tangent modulus :math:`\mathbb{C}` (=``dσdε``).

        ..  math::

            \mathbb{C} &= \lambda \ \boldsymbol{1} \otimes \boldsymbol{1} +
                2 \mu \ \boldsymbol{1} \odot \boldsymbol{1}

            \boldsymbol{\sigma} &= \boldsymbol{\sigma}_n
                + \mathbb{C} : \Delta\boldsymbol{\varepsilon}

    Examples
    --------
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.MaterialStrain(material=fem.linear_elastic, λ=2.0, μ=1.0)
        >>> ax = umat.plot()

    See Also
    --------
    MaterialStrain : A strain-based user-defined material definition with a given
        function for the stress tensor and the (fourth-order) elasticity tensor.

    """

    # change of stress due to change of strain
    eye = identity(dim=3, shape=(1, 1))
    dσ = 2 * μ * dε + λ * trace(dε) * eye

    # update stress
    σ = σn + dσ

    # evaluate elasticity tensor
    if kwargs["tangent"]:
        dσdε = 2 * μ * cdya_ik(eye, eye) + λ * dya(eye, eye)
    else:
        dσdε = None

    # update state variables (not used here)
    ζ = ζn

    return dσdε, σ, ζ



try:
    from . import models
    from ._helpers import isochoric_volumetric_split, vmap
    from ._hyperelastic import Hyperelastic
    from ._material import Material
    from ._total_lagrange import total_lagrange
    from ._updated_lagrange import updated_lagrange

    __all__ = [
        "Hyperelastic",
        "isochoric_volumetric_split",
        "Material",
        "models",
        "total_lagrange",
        "updated_lagrange",
        "vmap",
    ]

except ModuleNotFoundError:
    __all__ = []



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._helpers import mesh_or_data


def cell_types():
    r"""Return a list with tuples of identical cell types for different packages.

    Examples
    --------
    ..  pyvista-plot::
        :context:

        >>> import felupe as fem
        >>>
        >>> cell_types_array = fem.mesh.cell_types()

    Create a dict which takes a FElupe cell type string and returns the PyVista cell
    type.

    ..  pyvista-plot::
        :context:

        >>> cell_types_felupe_to_pyvista = dict(cell_types_array)
        >>> cell_types_felupe_to_pyvista["line"]
        <CellType.LINE: 3>

    Create a dict which takes a PyVista cell type string and returns the FElupe cell
    type.

    ..  pyvista-plot::
        :context:

        >>> import pyvista as pv
        >>>
        >>> cell_types_pyvista_to_felupe = dict(cell_types_array[:, [1, 0]])
        >>> cell_types_pyvista_to_felupe[pv.CellType.LINE]
        "line"
    """

    import pyvista as pv

    cell_types = [
        ("vertex", pv.CellType.VERTEX),
        ("line", pv.CellType.LINE),
        ("triangle", pv.CellType.TRIANGLE),
        ("triangle6", pv.CellType.QUADRATIC_TRIANGLE),
        ("tetra", pv.CellType.TETRA),
        ("tetra10", pv.CellType.QUADRATIC_TETRA),
        ("quad", pv.CellType.QUAD),
        ("quad8", pv.CellType.QUADRATIC_QUAD),
        ("quad9", pv.CellType.BIQUADRATIC_QUAD),
        ("hexahedron", pv.CellType.HEXAHEDRON),
        ("hexahedron20", pv.CellType.QUADRATIC_HEXAHEDRON),
        ("hexahedron27", pv.CellType.TRIQUADRATIC_HEXAHEDRON),
        ("VTK_LAGRANGE_HEXAHEDRON", pv.CellType.LAGRANGE_HEXAHEDRON),
        ("VTK_LAGRANGE_QUADRILATERAL", pv.CellType.LAGRANGE_QUADRILATERAL),
        ("VTK_LAGRANGE_LINE", pv.CellType.LAGRANGE_CURVE),
    ]

    return np.array(cell_types, dtype=object)


@mesh_or_data
def convert(
    points,
    cells,
    cell_type,
    order=0,
    calc_points=False,
    calc_midfaces=False,
    calc_midvolumes=False,
):
    """Convert a mesh to a given order. Only conversions to ``order=0`` and ``order=2``
    are supported. This function supports meshes with cell types ``"triangle"``,
    ``"tetra"``, ``"quad"`` and ``"hexahedron"``.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    cells : list or ndarray
        Original point-connectivity of cells.
    cell_type : str
        A string in VTK-convention that specifies the cell type. Must be one of
        ``"triangle"``, ``"tetra"``, ``"quad"`` or ``"hexahedron"``.
    order : int, optional
        The order of the converted mesh (default is 0). If 0, the points-array will be
        of shape ``(ncells, dim)``. If 0 and ``calc_points`` is True, the mean of all
        points per cell is evaluated. If 0 and ``calc_points`` is False, the points
        array is filled with zeros. If 2, at least midpoints on cell edges are added to
        the mesh. If 2 and ``calc_midfaces`` is True, midpoints on cell faces are also
        added. If 2 and ``calc_midvolumes`` is True, midpoints on cell volumes are also
        added. Raises an error if not 0 or 2.
    calc_points : bool, optional
        Flag to return the mean of all points per cell if ``order=0`` (default is
        False). If False, the points-array is filled with zeros.
    calc_midfaces : bool, optional
        Flag to add midpoints on cell faces if ``order=2`` (default is False).
    calc_midvolumes : bool, optional
        Flag to add midpoints on cell volumes if ``order=2`` (default is False).

    Returns
    -------
    points : ndarray
        Modified point coordinates.
    cells : list or ndarray
        Converted cells.
    cell_type : str or None
        A string in VTK-convention that specifies the cell type.

    Examples
    --------
    Convert a mesh of hexahedrons to quadratic hexahedrons by inserting midpoints on
    the cell edges.

    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.Rectangle(n=6)
       >>> mesh2 = fem.mesh.convert(mesh, order=2)
       >>>
       >>> mesh2.plot(plotter=mesh.plot(), style="points", color="black").show()

       >>> mesh2
       <felupe Mesh object>
         Number of points: 96
         Number of cells:
           quad8: 25

    See Also
    --------
    felupe.mesh.add_midpoints_edges : Add midpoints on edges for given points and cells
        and update cell_type accordingly.
    felupe.mesh.add_midpoints_faces : Add midpoints on faces for given points and cells
        and update cell_type accordingly.
    felupe.mesh.add_midpoints_volumes : Add midpoints on volumes for given points and
        cells and update cell_type accordingly.
    felupe.Mesh.add_midpoints_edges : Add midpoints on edges for given points and cells
        and update cell_type accordingly.
    felupe.Mesh.add_midpoints_faces : Add midpoints on faces for given points and cells
        and update cell_type accordingly.
    felupe.Mesh.add_midpoints_volumes : Add midpoints on volumes for given points and
        cells and update cell_type accordingly.

    """

    ncells = len(cells)
    dim = points.shape[1]

    if cell_type not in ["triangle", "tetra", "quad", "hexahedron"]:
        raise NotImplementedError("Cell type not supported for conversion.")

    if order == 0:
        if calc_points:
            points_new = np.stack([np.mean(points[cell], axis=0) for cell in cells])
        else:
            points_new = np.zeros((ncells, dim), dtype=int)

        cells_new = np.arange(ncells).reshape(-1, 1)
        cell_type_new = cell_type

    elif order == 2:
        points_new, cells_new, cell_type_new = add_midpoints_edges(
            points, cells, cell_type
        )

        if calc_midfaces:
            points_new, cells_new, cell_type_new = add_midpoints_faces(
                points_new, cells_new, cell_type_new
            )

        if calc_midvolumes:
            points_new, cells_new, cell_type_new = add_midpoints_volumes(
                points_new, cells_new, cell_type_new
            )

    else:
        raise NotImplementedError("Unsupported order conversion.")

    return points_new, cells_new, cell_type_new


@mesh_or_data
def collect_edges(points, cells, cell_type):
    """Collect all unique edges, calculate and return midpoints on edges as well as the
    additional cells array.

    See Also
    --------
    felupe.mesh.add_midpoints_edges : Add midpoints on cell edges for given points and
        cells and update cell_type accordingly.
    felupe.Mesh.add_midpoints_edges : Add midpoints on cell edges for given points and
        cells and update cell_type accordingly.
    """

    supported_cell_types = ["triangle", "tetra", "quad", "hexahedron"]

    if cell_type not in supported_cell_types:
        raise TypeError("Cell type not implemented.")

    number_of_edges = {"triangle": 3, "tetra": 6, "quad": 4, "hexahedron": 12}

    if cell_type in ["triangle", "tetra"]:
        # k-th edge is (i[k], j[k])
        i = [0, 1, 2, 3, 3, 3][: number_of_edges[cell_type]]
        j = [1, 2, 0, 0, 1, 2][: number_of_edges[cell_type]]

    elif cell_type in ["quad", "hexahedron"]:
        # k-th edge is (i[k], j[k])
        i = [0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3][: number_of_edges[cell_type]]
        j = [1, 2, 3, 0, 5, 6, 7, 4, 4, 5, 6, 7][: number_of_edges[cell_type]]

    edges_to_stack = cells[:, i], cells[:, j]

    # sort points of edges
    edges = np.sort(np.dstack(edges_to_stack).reshape(-1, 2), axis=1)

    # obtain unique edges and inverse mapping
    edges_unique, inverse = np.unique(edges, False, True, False, 0)

    # calculate midpoints on edges as mean
    points_edges = np.mean(points[edges_unique.T], axis=0)

    # create the additionals cells array
    cells_edges = inverse.reshape(len(cells), -1)

    return points_edges, cells_edges, cell_type


@mesh_or_data
def collect_faces(points, cells, cell_type):
    """Collect all unique faces, calculate and return midpoints on faces as well as the
    additional cells array.

    See Also
    --------
    felupe.mesh.add_midpoints_faces : Add midpoints on cell faces for given points and
        cells and update cell_type accordingly.
    felupe.Mesh.add_midpoints_faces : Add midpoints on cell faces for given points and
        cells and update cell_type accordingly.
    """

    supported_cell_types = [
        "triangle",
        "triangle6",
        "tetra",
        "tetra10",
        "quad",
        "quad8",
        "hexahedron",
        "hexahedron20",
    ]

    if cell_type not in supported_cell_types:
        raise TypeError("Cell type not implemented.")

    if "triangle" in cell_type:
        # k-th face is (i[k], j[k], k[k])
        i = [0]
        j = [1]
        k = [2]

        faces_to_stack = cells[:, i], cells[:, j], cells[:, k]

    if "tetra" in cell_type:
        # k-th face is (i[k], j[k], k[k])
        # ordering?
        i = [0, 0, 0, 1]
        j = [1, 1, 2, 2]
        k = [2, 3, 3, 3]

        faces_to_stack = cells[:, i], cells[:, j], cells[:, k]

    elif "quad" in cell_type:
        # k-th edge is (i[k], j[k], k[k], m[k])
        i = [0]
        j = [1]
        k = [2]
        m = [3]

        faces_to_stack = cells[:, i], cells[:, j], cells[:, k], cells[:, m]

    elif "hexahedron" in cell_type:
        # k-th edge is (i[k], j[k], k[k], l[k])
        i = [0, 1, 1, 2, 0, 4]
        j = [3, 2, 0, 3, 1, 5]
        k = [7, 6, 4, 7, 2, 6]
        m = [4, 5, 5, 6, 3, 7]

        faces_to_stack = cells[:, i], cells[:, j], cells[:, k], cells[:, m]

    # sort points of edges
    faces = np.sort(np.dstack(faces_to_stack).reshape(-1, len(faces_to_stack)), axis=1)

    # obtain unique edges and inverse mapping
    faces_unique, inverse = np.unique(faces, False, True, False, 0)

    # calculate midpoints on edges as mean
    points_faces = np.mean(points[faces_unique.T], axis=0)

    # create the additionals cells array
    cells_faces = inverse.reshape(len(cells), -1)

    return points_faces, cells_faces, cell_type


@mesh_or_data
def collect_volumes(points, cells, cell_type):
    """Collect all volumes, calculate and return midpoints on volumes as well as the
    additional cells array.

    See Also
    --------
    felupe.mesh.add_midpoints_volumes : Add midpoints on cell volumes for given points
        and cells and update cell_type accordingly.
    felupe.Mesh.add_midpoints_volumes : Add midpoints on cell volumes for given points
        and cells and update cell_type accordingly.

    """

    supported_cell_types = [
        "tetra",
        "tetra10",
        "tetra14",
        "hexahedron",
        "hexahedron20",
        "hexahedron26",
    ]

    if cell_type not in supported_cell_types:
        raise TypeError("Cell type not implemented.")

    if "tetra" in cell_type:
        number_of_vertices = 3

    elif "hexahedron" in cell_type:
        number_of_vertices = 8

    if cell_type in supported_cell_types:
        points_volumes = np.mean(points[cells][:, :number_of_vertices, :], axis=1)
        cells_volumes = np.arange(cells.shape[0]).reshape(-1, 1)

    return points_volumes, cells_volumes, cell_type


@mesh_or_data
def add_midpoints_edges(points, cells, cell_type, cell_type_new=None):
    """Add midpoints on edges for given points and cells and update cell_type
    accordingly.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    cells : list or ndarray
        Original point-connectivity of cells.
    cell_type : str
        A string in VTK-convention that specifies the cell type.
    cell_type_new: str or None, optional
        A string in VTK-convention that specifies the new cell type (default is None).
        If None, the cell type is chosen automatically.

    Returns
    -------
    points : ndarray
        Modified point coordinates.
    cells : ndarray
        Modified point-connectivity of cells.
    cell_type : str or None
        A string in VTK-convention that specifies the cell type.

    Examples
    --------
    Convert a mesh of hexahedrons to quadratic hexahedrons by inserting midpoints on
    the cell edges.

    .. pyvista-plot::
       :include-source: True
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.Rectangle(n=6)
       >>> mesh_with_midpoints_edges = fem.mesh.add_midpoints_edges(mesh)
       >>>
       >>> mesh_with_midpoints_edges.plot(
       ...     plotter=mesh.plot(), style="points", color="black"
       ... ).show()

    >>> mesh_with_midpoints_edges
    <felupe Mesh object>
      Number of points: 96
      Number of cells:
        quad8: 25

    See Also
    --------
    felupe.Mesh.add_midpoints_edges : Add midpoints on edges for given points and cells
        and update cell_type accordingly.

    """

    # collect edges
    points_edges, cells_edges, _ = collect_edges(
        points,
        cells,
        cell_type,
    )

    # add offset to point index for edge-midpoints
    # in additional cells array
    cells_edges += len(points)

    # vertical stack of points and horizontal stack of edges
    points_new = np.vstack((points, points_edges))
    cells_new = np.hstack((cells, cells_edges))

    if cell_type_new is None:
        cell_type_new = {
            "triangle": "triangle6",
            "tetra": "tetra10",
            "quad": "quad8",
            "hexahedron": "hexahedron20",
        }[cell_type]

    return points_new, cells_new, cell_type_new


@mesh_or_data
def add_midpoints_faces(points, cells, cell_type, cell_type_new=None):
    """Add midpoints on faces for given points and cells and update cell_type
    accordingly.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    cells : list or ndarray
        Original point-connectivity of cells.
    cell_type : str
        A string in VTK-convention that specifies the cell type.
    cell_type_new: str or None, optional
        A string in VTK-convention that specifies the new cell type (default is None).
        If None, the cell type is chosen automatically.

    Returns
    -------
    points : ndarray
        Modified point coordinates.
    cells : ndarray
        Modified point-connectivity of cells.
    cell_type : str or None
        A string in VTK-convention that specifies the cell type.

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>>
       >>> mesh = fem.Rectangle(n=6)
       >>> mesh_with_midpoints_faces = fem.mesh.add_midpoints_faces(
       ...     mesh, cell_type_new="quad"
       ... )
       >>>
       >>> plotter = mesh_with_midpoints_faces.plot(
       ...     plotter=mesh.plot(), style="points", color="black"
       ... ).show()

    >>> mesh_with_midpoints_faces
    <felupe Mesh object>
      Number of points: 61
      Number of cells:
        quad: 25

    See Also
    --------
    felupe.Mesh.add_midpoints_faces : Add midpoints on faces for given points and cells
        and update cell_type accordingly.
    """

    # collect faces
    points_faces, cells_faces, _ = collect_faces(
        points,
        cells,
        cell_type,
    )

    # add offset to point index for faces-midpoints
    # in additional cells array
    cells_faces += len(points)

    # vertical stack of points and horizontal stack of edges
    points_new = np.vstack((points, points_faces))
    cells_new = np.hstack((cells, cells_faces))

    if cell_type_new is None:
        cell_type_new = {
            None: None,
            "triangle": None,
            "triangle6": "triangle7",
            "tetra10": "tetra14",
            "quad": None,
            "quad8": "quad9",
            "hexahedron": None,
            "hexahedron20": "hexahedron26",
        }[cell_type]

    return points_new, cells_new, cell_type_new


@mesh_or_data
def add_midpoints_volumes(points, cells, cell_type, cell_type_new=None):
    """Add midpoints on volumes for given points and cells and update cell_type
    accordingly.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    cells : list or ndarray
        Original point-connectivity of cells.
    cell_type : str
        A string in VTK-convention that specifies the cell type.
    cell_type_new: str or None, optional
        A string in VTK-convention that specifies the new cell type (default is None).
        If None, the cell type is chosen automatically.

    Returns
    -------
    points : ndarray
        Modified point coordinates.
    cells : ndarray
        Modified point-connectivity of cells.
    cell_type : str or None
        A string in VTK-convention that specifies the cell type.

    Examples
    --------
    .. pyvista-plot::
       :force_static:

       >>> import felupe as fem
       >>> import pyvista as pv
       >>>
       >>> mesh = fem.Cube(n=6)
       >>> mesh_with_midpoints_volumes = fem.mesh.add_midpoints_volumes(
       ...     mesh, cell_type_new="hexahedron9"
       ... )
       >>> plotter = pv.Plotter()
       >>> actor = plotter.add_points(mesh_with_midpoints_volumes.points, color="black")
       >>> mesh.plot(opacity=0.5, plotter=plotter).show()

    >>> mesh_with_midpoints_volumes
    <felupe Mesh object>
      Number of points: 341
      Number of cells:
        hexahedron9: 125

    See Also
    --------
    felupe.Mesh.add_midpoints_volumes : Add midpoints on volumes for given points and
        cells and update cell_type accordingly.
    """

    # collect volumes
    points_volumes, cells_volumes, _ = collect_volumes(
        points,
        cells,
        cell_type,
    )

    # add offset to point index for volumes-midpoints
    # in additional cells array
    cells_volumes += len(points)

    # vertical stack of points and horizontal stack of edges
    points_new = np.vstack((points, points_volumes))
    cells_new = np.hstack((cells, cells_volumes))

    if cell_type_new is None:
        cell_type_new = {
            None: None,
            "tetra": None,
            "tetra10": None,
            "tetra14": "tetra15",
            "hexahedron": None,
            "hexahedron20": None,
            "hexahedron26": "hexahedron27",
        }[cell_type]

    return points_new, cells_new, cell_type_new


@mesh_or_data
def subdivide(points, cells, cell_type):
    """Subdivide cells by adding midpoints on edges, faces and volumes.

    Parameters
    ----------
    points : list or ndarray
        Original point coordinates.
    cells : list or ndarray
        Original point-connectivity of cells.
    cell_type : str
        A string in VTK-convention that specifies the cell type.

    Returns
    -------
    points : ndarray
        Modified point coordinates.
    cells : ndarray
        Modified point-connectivity of cells.
    cell_type : str or None
        A string in VTK-convention that specifies the cell type.

    Examples
    --------
    Take a rectangle mesh and subdivide it two times.

    .. pyvista-plot::
       :context:
       :force_static:

       >>> import felupe as fem
       >>>
       >>> rect = fem.Rectangle(n=4).modify_corners()
       >>> rect.plot().show()

    .. pyvista-plot::
       :context:
       :force_static:

       >>> mesh = fem.mesh.subdivide(fem.mesh.subdivide(rect))
       >>> mesh.plot().show()

    See Also
    --------
    felupe.Mesh.subdivide : Subdivide cells by adding midpoints on edges, faces and
        volumes.
    """

    if cell_type == "triangle":
        points, cells, cell_type_new = add_midpoints_edges(points, cells, cell_type)

        idx = [[0, 3, 5], [3, 1, 4], [5, 4, 2], [3, 4, 5]]

    elif cell_type == "tetra":
        points, cells, cell_type_new = add_midpoints_edges(points, cells, cell_type)
        points, cells, cell_type_new = add_midpoints_faces(points, cells, cell_type_new)

        idx = [
            [0, 4, 6, 7],
            [4, 1, 5, 8],
            [6, 5, 2, 9],
            [7, 8, 9, 3],
            [8, 4, 5, 9],
            [4, 5, 6, 9],
            [8, 4, 7, 9],
            [7, 6, 9, 4],
        ]

    elif cell_type == "quad":
        points, cells, cell_type_new = add_midpoints_edges(points, cells, cell_type)
        points, cells, cell_type_new = add_midpoints_faces(points, cells, cell_type_new)

        idx = [[0, 4, 8, 7], [4, 1, 5, 8], [7, 8, 6, 3], [8, 5, 2, 6]]

    elif cell_type == "hexahedron":
        points, cells, cell_type_new = add_midpoints_edges(points, cells, cell_type)
        points, cells, cell_type_new = add_midpoints_faces(points, cells, cell_type_new)
        points, cells, cell_type_new = add_midpoints_volumes(
            points, cells, cell_type_new
        )

        idx = [
            [0, 8, 24, 11, 16, 22, 26, 20],
            [8, 1, 9, 24, 22, 17, 21, 26],
            [11, 24, 10, 3, 20, 26, 23, 19],
            [24, 9, 2, 10, 26, 21, 18, 23],
            [16, 22, 26, 20, 4, 12, 25, 15],
            [22, 17, 21, 26, 12, 5, 13, 25],
            [20, 26, 23, 19, 15, 25, 14, 7],
            [26, 21, 18, 23, 25, 13, 6, 14],
        ]

    else:
        raise NotImplementedError("Cell type not supported for subdivision.")

    cells = np.vstack([cells[:, i] for i in idx])

    return points, cells, cell_type
